# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause

cdef void _matrix_profile_stmp(
    const double *x,
    Py_ssize_t x_length,
    const double *y,
    Py_ssize_t y_length,
    Py_ssize_t window,
    Py_ssize_t exclude,
    double *mean_x,
    double *std_x,
    complex *x_buffer,
    complex *y_buffer,
    double *dist_buffer,
    double *mp,
    Py_ssize_t *mpi,
) nogil


# Authors: Isak Samsten
# License: BSD 3 clause

from functools import wraps

import numpy as np
from sklearn.utils.validation import _is_arraylike_not_scalar


def _array_or_scalar(x, squeeze=True):
    if not isinstance(x, np.ndarray):
        return x

    if x.size == 1:
        return x.item()
    else:
        return np.squeeze(x) if squeeze else x


# noqa: H0002
def array_or_scalar(optional_f=None, squeeze=True):
    """
    Decorate a function returning.

    An ndarray to return a single scalar if the array has a single item.

    Parameters
    ----------
    optional_f : callable, optional
        Used if the decorator is used without arguments.
    squeeze : bool, optional
        Remove axis of length one from the returned arrays.

    Returns
    -------
    function
        Decorated function.
    """

    def decorator(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            x = f(*args, **kwargs)
            if isinstance(x, tuple):
                return tuple(_array_or_scalar(r, squeeze=squeeze) for r in x)
            elif isinstance(x, np.ndarray):
                return _array_or_scalar(x, squeeze=squeeze)
            else:
                return x

        return wrap

    return decorator(optional_f) if callable(optional_f) else decorator


def _singleton(x):
    if _is_arraylike_not_scalar(x):
        if isinstance(x, np.ndarray) and x.size == 1:
            return x.item()
        else:
            return x[0] if len(x) == 1 else x
    else:
        return x


# noqa: H0002
def singleton(f):
    """
    Recursivley try to unwrap list return arguments.

    Such that a single element can be returned.

    Parameters
    ----------
    f : function
        The function to wrap.

    Returns
    -------
    function
        The wrapper.
    """

    @wraps(f)
    def wrap(*args, **kwargs):
        ret_vals = f(*args, **kwargs)
        if isinstance(ret_vals, tuple):
            return tuple(_singleton(ret_val) for ret_val in ret_vals)
        else:
            return _singleton(ret_vals)

    return wrap


# noqa: H0002
def unstable(optional_f=None, stability="beta", description=None):
    """
    Decorate a function as unsatable.

    Parameters
    ----------
    optional_f : callable, optional
        The decorated function.
    stability : str, optional
        The stability of the feature.
    description : str, optional
        The description of the feature.

    Returns
    -------
    function
        Decorated function.
    """
    import warnings

    def decorator(f):
        @wraps(f)
        def wrap(*args, **kwargs):
            warning = f"'{f.__qualname__}' is currently in '{stability}'."
            if description is not None:
                warning += f" {description}."
            warnings.warn(warning, FutureWarning)
            return f(*args, **kwargs)

        return wrap

    return decorator(optional_f) if callable(optional_f) else decorator



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause
from numpy cimport float64_t, ndarray

from ..utils cimport TSArray
from ..utils._misc cimport List

# Threshold for when standard deviation is considered to be zero.
cdef double EPSILON = 1e-13

cdef class MetricList(List):

    cdef int reset(self, Py_ssize_t metric, TSArray X, TSArray Y) noexcept nogil

    cdef double distance(
        self,
        Py_ssize_t metric,
        TSArray X,
        Py_ssize_t x_index,
        TSArray Y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
    ) noexcept nogil

    cdef double _distance(
        self,
        Py_ssize_t metric,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil

cdef class SubsequenceMetricList(List):

    cdef int reset(self, Py_ssize_t metric, TSArray X) noexcept nogil

    cdef int init_transient(
        self,
        Py_ssize_t metric,
        TSArray X,
        SubsequenceView *v,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) noexcept nogil

    cdef int init_persistent(
        self,
        Py_ssize_t metric,
        TSArray X,
        SubsequenceView* v,
        Subsequence* s,
    ) noexcept nogil

    cdef int free_transient(self, Py_ssize_t metric, SubsequenceView *t) noexcept nogil

    cdef int free_persistent(self, Py_ssize_t metric, Subsequence *t) noexcept nogil

    cdef int from_array(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        object obj,
    )

    cdef object to_array(
        self,
        Py_ssize_t metric,
        Subsequence *s
    )

    cdef double transient_distance(
        self,
        Py_ssize_t metric,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index,
    ) noexcept nogil

    cdef double persistent_distance(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index,
    ) noexcept nogil

    cdef Py_ssize_t transient_matches(
        self,
        Py_ssize_t metric,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indices,
    ) noexcept nogil

    cdef Py_ssize_t persistent_matches(
        self,
        Py_ssize_t metric,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indices,
    ) noexcept nogil


cdef struct SubsequenceView:
    Py_ssize_t index  # the index of the shapelet sample
    Py_ssize_t start  # the start position
    Py_ssize_t length  # the length of the shapelet
    Py_ssize_t dim  # the dimension of the shapelet
    double mean  # the mean of the shapelet
    double std  # the stanard devision
    void *extra


cdef struct Subsequence:
    Py_ssize_t length
    Py_ssize_t dim
    double mean
    double std
    int ts_index
    int ts_start
    double *data
    void *extra

cdef class SubsequenceMetric:

    cdef int reset(self, TSArray X) noexcept nogil

    cdef int init_transient(
        self,
        TSArray X,
        SubsequenceView *v,
        Py_ssize_t index,
        Py_ssize_t start,
        Py_ssize_t length,
        Py_ssize_t dim,
    ) noexcept nogil

    cdef int init_persistent(
        self,
        TSArray X,
        SubsequenceView* v,
        Subsequence* s,
    ) noexcept nogil

    cdef int free_transient(self, SubsequenceView *t) noexcept nogil

    cdef int free_persistent(self, Subsequence *t) noexcept nogil

    cdef int from_array(
        self,
        Subsequence *s,
        object obj,
    )

    cdef object to_array(
        self,
        Subsequence *s
    )

    cdef double transient_distance(
        self,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index=*,
    ) noexcept nogil

    cdef double persistent_distance(
        self,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        Py_ssize_t *return_index=*,
    ) noexcept nogil

    cdef Py_ssize_t transient_matches(
        self,
        SubsequenceView *v,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil

    cdef Py_ssize_t persistent_matches(
        self,
        Subsequence *s,
        TSArray X,
        Py_ssize_t index,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil

    cdef void transient_profile(
        self,
        SubsequenceView *s,
        TSArray x,
        Py_ssize_t i,
        double *dp,
    ) noexcept nogil

    cdef void persistent_profile(
        self,
        Subsequence *s,
        TSArray x,
        Py_ssize_t i,
        double *dp,
    ) noexcept nogil

    cdef void _distance_profile(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double *dp,
    ) noexcept nogil

    cdef double _distance(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        Py_ssize_t *return_index=*,
    ) noexcept nogil

    cdef Py_ssize_t _matches(
        self,
        const double *s,
        Py_ssize_t s_len,
        double s_mean,
        double s_std,
        void *s_extra,
        const double *x,
        Py_ssize_t x_len,
        double threshold,
        double *distances,
        Py_ssize_t *indicies,
    ) noexcept nogil


cdef class ScaledSubsequenceMetric(SubsequenceMetric):
    pass


cdef class Metric:

    cdef int reset(self, TSArray X, TSArray Y) noexcept nogil

    cdef double distance(
        self,
        TSArray X,
        Py_ssize_t x_index,
        TSArray Y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
    ) noexcept nogil

    cdef double _distance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len
    ) noexcept nogil

    # Compute the distance between X[i] and Y[j].
    # If possible, the computation should be aborted if
    # the distance >= lower_bound. To disable the upper bound
    # one can pass INFINITY.
    #
    # `lower_bound` has the initial best-so-far distance on entry.
    #  - If distance < lower_bound, `lower_bound` is the actual distance on
    #    exit and True is returned.
    #  - Otherwise, `lower_bound` is unchanged and the return value is
    #    False
    cdef bint eadistance(
        self,
        TSArray X,
        Py_ssize_t x_index,
        TSArray Y,
        Py_ssize_t y_index,
        Py_ssize_t dim,
        double *lower_bound,
    ) noexcept nogil

    cdef bint _eadistance(
        self,
        const double *x,
        Py_ssize_t x_len,
        const double *y,
        Py_ssize_t y_len,
        double *upper_bound,
    ) noexcept nogil


cdef Py_ssize_t dilated_distance_profile(
    Py_ssize_t stride,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double *kernel,
    Py_ssize_t k_len,
    const double* x,
    Py_ssize_t x_len,
    Metric metric,
    double *x_buffer,
    double *k_buffer,
    double ea,
    double* out,
) noexcept nogil

cdef Py_ssize_t scaled_dilated_distance_profile(
    Py_ssize_t stride,
    Py_ssize_t dilation,
    Py_ssize_t padding,
    double *kernel,
    Py_ssize_t k_len,
    const double* x,
    Py_ssize_t x_len,
    Metric metric,
    double *x_buffer,
    double *k_buffer,
    double ea,
    double* out,
) noexcept nogil



# Authors: Isak Samsten
# License: BSD 3 clause

import numbers
import warnings

import numpy as np
from sklearn.utils._param_validation import Interval, StrOptions

from wildboar.utils.validation import check_option

from ..distance._distance import _METRICS
from ..distance._multi_metric import make_metrics
from ..tree._cptree import (
    EntropyCriterion,
    GiniCriterion,
    LabelPivotSampler,
    Tree,
    TreeBuilder,
    UniformMetricSampler,
    UniformPivotSampler,
    WeightedMetricSampler,
)
from ._base import BaseTree, BaseTreeClassifier

_CLF_CRITERION = {
    "gini": GiniCriterion,
    "entropy": EntropyCriterion,
}

_PIVOT_SAMPLER = {
    "label": LabelPivotSampler,
    "uniform": UniformPivotSampler,
}

_METRICS_SAMPLER = {
    "uniform": UniformMetricSampler,
    "weighted": WeightedMetricSampler,
}

_METRIC_NAMES = set(_METRICS.keys())
_METRIC_NAMES.add("auto")
_METRIC_NAMES.add("default")  # TODO(1.3)
_METRIC_NAMES = frozenset(_METRIC_NAMES)


# noqa: D409
class ProximityTreeClassifier(BaseTreeClassifier):
    """
    A classifier that uses a k-branching tree based on pivot-time series.

    Parameters
    ----------
    n_pivot : int, optional
        The number of pivots to sample at each node.
    criterion : {"entropy", "gini"}, optional
        The impurity criterion.
    pivot_sample : {"label", "uniform"}, optional
        The pivot sampling method.
    metric_sample : {"uniform", "weighted"}, optional
        The metric sampling method.
    metric : {"auto", "default"}, str or list, optional
        The distance metrics. By default, we use the parameterization suggested by
        Lucas et.al (2019).

        - If "auto", use the default metric specification, suggested by
          (Lucas et. al, 2020).
        - If str, use a single metric or default metric specification.
        - If list, custom metric specification can be given as a list of
          tuples, where the first element of the tuple is a metric name and the
          second element a dictionary with a parameter grid specification. A
          parameter grid specification is a `dict` with two mandatory and one
          optional key-value pairs defining the lower and upper bound on the
          values as well as the number of values in the grid. For example, to
          specifiy a grid over the argument 'r' with 10 values in the range 0
          to 1, we would give the following specification: 
          `dict(min_r=0, max_r=1, num_r=10)`.

        Read more about the metrics and their parameters in the
        :ref:`User guide <list_of_metrics>`.
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the :ref:`User guide
        <list_of_metrics>`.
    metric_factories : dict, optional
        A metric specification.

        .. deprecated:: 1.2
            Use the combination of metric and metric params.
    max_depth : int, optional
        The maximum tree depth.
    min_samples_split : int, optional
        The minimum number of samples to consider a split.
    min_samples_leaf : int, optional
        The minimum number of samples in a leaf.
    min_impurity_decrease : float, optional
        The minimum impurity decrease to build a sub-tree.
    class_weight : dict or "balanced", optional
        Weights associated with the labels.

        - if dict, weights on the form {label: weight}.
        - if "balanced" each class weight inversely proportional to the class
            frequency.
        - if None, each class has equal weight.
    random_state : int or RandomState
        - If `int`, `random_state` is the seed used by the random number generator
        - If `RandomState` instance, `random_state` is the random number generator
        - If `None`, the random number generator is the `RandomState` instance used
            by `np.random`.

    References
    ----------
    Lucas, Benjamin, Ahmed Shifaz, Charlotte Pelletier, Lachlan O'Neill, Nayyar Zaidi, \
    Bart Goethals, François Petitjean, and Geoffrey I. Webb. (2019)
        Proximity forest: an effective and scalable distance-based classifier for time
        series. Data Mining and Knowledge Discovery

    Examples
    --------
    Fit a single proximity tree, with dynamic time warping and move-split-merge metrics.

    >>> from wildboar.datasets import load_dataset
    >>> from wildboar.tree import ProximityTreeClassifier
    >>> x, y = load_dataset("GunPoint")
    >>> f = ProximityTreeClassifier(
    ...     n_pivot=10,
    ...     metrics=[
    ...         ("dtw", {"min_r": 0.1, "max_r": 0.25}),
    ...         ("msm", {"min_c": 0.1, "max_c": 100, "num_c": 20})
    ...     ],
    ...     criterion="gini"
    ... )
    >>> f.fit(x, y)

    """

    _parameter_constraints: dict = {
        **BaseTree._parameter_constraints,
        "n_pivot": [
            Interval(numbers.Integral, 1, None, closed="left"),
        ],
        "criterion": [
            StrOptions(_CLF_CRITERION.keys()),
        ],
        "pivot_sample": [
            StrOptions(_PIVOT_SAMPLER.keys()),
        ],
        "metric_sample": [
            StrOptions(_METRICS_SAMPLER.keys()),
        ],
        "metric": [
            StrOptions(_METRIC_NAMES, deprecated={"default"}),
            list,
        ],
        "metric_params": [None, dict],
        "metric_factories": [  # TODO(1.4)
            StrOptions({"default", "auto"}, deprecated={"default"}),
            list,
            None,
        ],
        "random_state": ["random_state"],
        "class_weight": [
            StrOptions({"balanced"}),
            dict,
            None,
        ],
    }

    def __init__(
        self,
        n_pivot=1,
        *,
        criterion="entropy",
        pivot_sample="label",
        metric_sample="weighted",
        metric="auto",
        metric_params=None,
        metric_factories=None,
        max_depth=None,
        min_samples_split=2,
        min_samples_leaf=1,
        min_impurity_decrease=0.0,
        class_weight=None,
        random_state=None,
    ):
        super().__init__(
            max_depth=max_depth,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_impurity_decrease=min_impurity_decrease,
        )
        self.n_pivot = n_pivot
        self.criterion = criterion
        self.pivot_sample = pivot_sample
        self.metric_sample = metric_sample
        self.metric = metric
        self.metric_params = metric_params
        self.metric_factories = metric_factories
        self.class_weight = class_weight
        self.random_state = random_state

    def _fit(self, x, y, sample_weights, max_depth, random_state):
        if self.metric_factories is not None:
            # TODO(1.4)
            warnings.warn(
                "The parameter metric_factories has been renamed to metric "
                "in 1.2 and will be removed in 1.4",
                DeprecationWarning,
            )
            if isinstance(self.metric_factories, dict):
                metric = list(self.metric_factories.items())

            metric = self.metric_factories

        metric = self.metric
        if isinstance(metric, str) and metric in ["default", "auto"]:
            if metric == "default":
                # TODO(1.4)
                warnings.warn(
                    "The default value for 'metric' has changed to 'auto' in 1.2 and "
                    "'default' will be removed in 1.4.",
                    DeprecationWarning,
                )

            std_x = np.std(x)
            metric_spec = [
                ("euclidean", None),
                ("dtw", None),
                ("ddtw", None),
                ("dtw", {"min_r": 0, "max_r": 0.25, "default_n": 50}),
                ("ddtw", {"min_r": 0, "max_r": 0.25, "default_n": 50}),
                ("wdtw", {"min_g": std_x * 0.2, "max_g": std_x, "default_n": 50}),
                ("wddtw", {"min_g": std_x * 0.2, "max_g": std_x, "default_n": 50}),
                (
                    "lcss",
                    {
                        "min_epsilon": std_x * 0.2,
                        "max_epsilon": std_x,
                        "min_r": 0,
                        "max_r": 0.25,
                        "default_n": 20,
                    },
                ),
                ("erp", {"min_g": 0, "max_g": 1.0, "default_n": 50}),
                ("msm", {"min_c": 0.01, "max_c": 100, "default_n": 50}),
                (
                    "twe",
                    {
                        "min_penalty": 0.00001,
                        "max_penalty": 1.0,
                        "min_stiffness": 0.000001,
                        "max_stiffness": 0.1,
                        "default_n": 20,
                    },
                ),
            ]
            metrics, weights = make_metrics(metric_spec)
        elif isinstance(metric, str) and metric in _METRIC_NAMES:
            Metric = check_option(_METRICS, metric, "metric")
            metrics = [Metric(**(self.metric_params or {}))]
            weights = np.ones(1)
        else:
            metrics, weights = make_metrics(metric)

        Criterion = _CLF_CRITERION[self.criterion]
        PivotSampler = _PIVOT_SAMPLER[self.pivot_sample]
        MetricSampler = _METRICS_SAMPLER[self.metric_sample]
        tree_builder = TreeBuilder(
            x,
            sample_weights,
            PivotSampler(),
            MetricSampler(len(metrics), weights=weights),
            Criterion(y, self.n_classes_),
            Tree(metrics, self.n_classes_),
            random_state,
            n_attributes=self.n_pivot,
            max_depth=max_depth,
            min_impurity_decrease=self.min_impurity_decrease,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
        )
        tree_builder.build_tree()
        self.tree_ = tree_builder.tree



import math
import numbers
import warnings

import numpy as np
from sklearn.base import ClassifierMixin, ClusterMixin, TransformerMixin
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils._param_validation import Interval, StrOptions
from sklearn.utils.validation import check_is_fitted, check_random_state

from ..base import BaseEstimator
from ..utils.validation import check_classification_targets
from ._cneighbors import _pam_build, _pam_optimal_swap
from ._distance import _METRICS, argmin_distance, paired_distance, pairwise_distance
from .dtw import dtw_average


class KNeighborsClassifier(ClassifierMixin, BaseEstimator):
    """
    Classifier implementing k-nearest neighbors.

    Parameters
    ----------
    n_neighbors : int, optional
        The number of neighbors.
    metric : str, optional
        The distance metric.
    metric_params : dict, optional
        Optional parameters to the distance metric.

        Read more about the metrics and their parameters in the
        :ref:`User guide <list_of_metrics>`.
    n_jobs : int, optional
        The number of parallel jobs.

    Attributes
    ----------
    classes_ : ndarray of shapel (n_classes, )
        Known class labels.
    """

    _parameter_constraints: dict = {
        "n_neighbors": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [StrOptions(_METRICS.keys())],
        "metric_params": [dict, None],
        "n_jobs": [numbers.Integral, None],
    }

    def __init__(
        self,
        n_neighbors=5,
        *,
        metric="euclidean",
        metric_params=None,
        n_jobs=None,
    ):
        self.n_neighbors = n_neighbors
        self.metric = metric
        self.metric_params = metric_params
        self.n_jobs = n_jobs

    def fit(self, x, y):
        """
        Fit the classifier to the training data.

        Parameters
        ----------
        x : univariate time-series or multivaraite time-series
            The input samples.
        y : array-like of shape (n_samples, )
            The input labels.

        Returns
        -------
        KNeighborClassifier
            This instance.
        """
        self._validate_params()
        x, y = self._validate_data(x, y, allow_3d=True)
        check_classification_targets(y)
        self._fit_X = x.copy()  # Align naming with sklearn
        self.classes_, self._y = np.unique(y, return_inverse=True)
        return self

    def predict_proba(self, x):
        """
        Compute probability estimates for the samples in x.

        Parameters
        ----------
        x : univariate time-series or multivariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, len(self.classes_))
            The probability of each class for each sample.
        """
        check_is_fitted(self)
        x = self._validate_data(x, allow_3d=True, reset=False)

        # Treat a multivariate time series with a single dimension as a
        # univariate time series to ensure that we use the fast path.
        if x.ndim == 3 and x.shape[1] == 1:
            x = x.reshape(x.shape[0], -1)

        if x.ndim == 3:
            dists = pairwise_distance(
                x,
                self._fit_X,
                dim="mean",
                metric=self.metric,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs,
            )

            closest = np.argpartition(dists, self.n_neighbors, axis=1)[
                :, : self.n_neighbors
            ]
        else:
            closest = argmin_distance(
                x,
                self._fit_X,
                metric=self.metric,
                metric_params=self.metric_params,
                k=self.n_neighbors,
                n_jobs=self.n_jobs,
            )

        preds = self._y[closest]
        probs = np.empty((x.shape[0], len(self.classes_)), dtype=float)
        for i in range(len(self.classes_)):
            probs[:, i] = np.sum(preds == i, axis=1) / self.n_neighbors
        return probs

    def predict(self, x):
        """
        Compute the class label for the samples in x.

        Parameters
        ----------
        x : univariate time-series or multivariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, )
            The class label for each sample.
        """
        proba = np.argmax(self.predict_proba(x), axis=1)
        return np.take(self.classes_, proba)


class _KMeansCluster:
    def __init__(self, init, metric, metric_params, random_state):
        self.metric = metric
        self.metric_params = metric_params
        self.centroids_ = init
        self.random_state = random_state

    def cost(self, x):
        if not hasattr(self, "assigned_") or self.assigned_ is None:
            raise ValueError("`assign` must be called before `cost`")

        return (
            paired_distance(
                x,
                self.centroids_[self.assigned_],
                dim="mean",
                metric=self.metric,
                metric_params=self.metric_params,
            ).sum()
            / x.shape[0]
        )

    def assign(self, x):
        if x.ndim == 2:
            self.assigned_, self.distance_ = argmin_distance(
                x,
                self.centroids_,
                k=1,
                metric=self.metric,
                metric_params=self.metric_params,
                return_distance=True,
            )
            self.assigned_ = np.ravel(self.assigned_)
        else:
            self.distance_ = pairwise_distance(
                x,
                self.centroids_,
                dim="mean",
                metric=self.metric,
                metric_params=self.metric_params,
            )
            self.assigned_ = self.distance_.argmin(axis=1)

    def update(self, x):
        for c in range(self.centroids_.shape[0]):
            cluster = x[self.assigned_ == c]
            if cluster.shape[0] == 0:
                far_from_center = self.distance_.min(axis=1).argmax()
                self.assigned_[far_from_center] = c
                self.centroids_[c] = x[far_from_center]
            elif cluster.shape[0] == 1:
                self.centroids_[c] = cluster
            else:
                self.centroids_[c] = self._update_centroid(cluster, self.centroids_[c])


class _EuclideanCluster(_KMeansCluster):
    def __init__(self, centroids, random_state):
        super().__init__(centroids, "euclidean", None, random_state)

    def _update_centroid(self, group, _centroid):
        return group.mean(axis=0)


class _DtwCluster(_KMeansCluster):
    def __init__(self, centroids, r, random_state):
        super().__init__(centroids, "dtw", {"r": r}, random_state)
        self.r = r

    def _update_centroid(self, group, current_center):
        return dtw_average(
            group,
            r=self.r,
            init=current_center,
            method="mm",
            random_state=self.random_state.randint(np.iinfo(np.int32).max),
        )


class _WDtwCluster(_KMeansCluster):
    def __init__(self, centroids, r, g, random_state):
        super().__init__(centroids, "wdtw", {"r": r, "g": g}, random_state)
        self.r = r
        self.g = g

    def _update_centroid(self, group, current_center):
        return dtw_average(
            group,
            r=self.r,
            g=self.g,
            init=current_center,
            method="mm",
            random_state=self.random_state.randint(np.iinfo(np.int32).max),
        )


class KMeans(ClusterMixin, TransformerMixin, BaseEstimator):  # noqa: I0012
    """
    KMeans clustering with support for DTW and weighted DTW.

    Parameters
    ----------
    n_clusters : int, optional
        The number of clusters.
    metric : {"euclidean", "dtw"}, optional
        The metric.
    r : float, optional
        The size of the warping window.
    g : float, optional
        SoftDTW penalty. If None, traditional DTW is used.
    init : {"random"}, optional
        Cluster initialization. If "random", randomly initialize `n_clusters`.
    n_init : "auto" or int, optional
        Number times the algorithm is re-initialized with new centroids.
    max_iter : int, optional
        The maximum number of iterations for a single run of the algorithm.
    tol : float, optional
        Relative tolerance to declare convergence of two consecutive iterations.
    verbose : int, optional
        Print diagnostic messages during convergence.
    random_state : RandomState or int, optional
        Determines random number generation for centroid initialization and
        barycentering when fitting with `metric="dtw"`.

    Attributes
    ----------
    n_iter_ : int
        The number of iterations before convergence.
    cluster_centers_ : ndarray of shape (n_clusters, n_timestep)
        The cluster centers.
    labels_ : ndarray of shape (n_samples, )
        The cluster assignment.
    """

    _parameter_constraints: dict = {
        "n_clusters": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [StrOptions({"euclidean", "dtw"})],
        "r": [Interval(numbers.Real, 0, 1, closed="both")],
        "g": [None, Interval(numbers.Real, 0, None, closed="neither")],
        "init": [StrOptions({"random"})],
        "n_init": [
            StrOptions({"auto"}),
            Interval(numbers.Integral, 1, None, closed="left"),
        ],
        "max_iter": [Interval(numbers.Integral, 1, None, closed="left")],
        "tol": [float],
        "verbose": [int],
        "random_state": ["random_state", None],
    }

    def __init__(
        self,
        n_clusters=8,
        *,
        metric="euclidean",
        r=1.0,
        g=None,
        init="random",
        n_init="auto",
        max_iter=300,
        tol=1e-3,
        verbose=0,
        random_state=None,
    ):
        self.metric = metric
        self.r = r
        self.g = g
        self.n_clusters = n_clusters
        self.init = init
        self.n_init = n_init
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.random_state = random_state

    def fit(self, x, y=None):
        """
        Compute the kmeans-clustering.

        Parameters
        ----------
        x : univariate time-series
            The input samples.
        y : Ignored, optional
            Not used.

        Returns
        -------
        object
            Fitted estimator.
        """
        self._validate_params()
        x = self._validate_data(x, allow_3d=False)
        if self.n_init == "auto":
            n_init = 1
        else:
            n_init = self.n_init

        random_state = check_random_state(self.random_state)
        best_cost = np.inf
        best_cluster = None
        best_reassign = True
        best_iter = None
        for init in range(n_init):
            iter, cost, cluster, reassign = self._fit_one_init(x, random_state)
            if cost < best_cost:
                best_cost = cost
                best_cluster = cluster
                best_reassign = reassign
                best_iter = iter

        self.n_iter_ = best_iter
        self.inertia_ = best_cost
        self.cluster_centers_ = best_cluster.centroids_

        if best_reassign:
            best_cluster.assign(x)

        self.labels_ = best_cluster.assigned_
        return self

    def _fit_one_init(self, x, random_state):
        if self.init == "random":
            centroids = x[
                random_state.choice(x.shape[0], size=self.n_clusters, replace=False)
            ]
        else:
            raise ValueError(f"Unsupported 'init', got {self.init}")

        if self.metric == "euclidean":
            cluster = _EuclideanCluster(centroids, random_state=random_state)
        elif self.metric == "dtw":
            if self.g is None:
                cluster = _DtwCluster(centroids, r=self.r, random_state=random_state)
            else:
                cluster = _WDtwCluster(
                    centroids, r=self.r, g=self.g, random_state=random_state
                )
        else:
            raise ValueError(f"Unsupported 'metric', got {self.metric}")

        prev_cost = np.inf
        cost = -np.inf
        reassign = True
        for iter in range(self.max_iter):
            cluster.assign(x)
            prev_cost, cost = cost, cluster.cost(x)
            if self.verbose > 0:
                print(f"Iteration {iter}, {cost} (prev_cost = {prev_cost})")

            if math.isclose(cost, prev_cost, rel_tol=self.tol):
                reassign = False
                break

            cluster.update(x)

        return iter, cost, cluster, reassign

    def predict(self, x):
        """
        Predict the closest cluster for each sample.

        Parameters
        ----------
        x : univariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, )
            Index of the cluster each sample belongs to.
        """
        return self.transform(x).argmin(axis=1)

    def transform(self, x):
        """
        Transform the input to a cluster distance space.

        Parameters
        ----------
        x : univariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, n_clusters)
            The distance between each sample and each cluster.
        """
        check_is_fitted(self)
        self._validate_data(x, allow_3d=False, reset=False)
        metric_params = {}
        metric = self.metric
        if self.metric == "dtw":
            metric_params["r"] = self.r
            if self.g is not None:
                metric_params["g"] = self.g
                metric = "wdtw"

        return pairwise_distance(
            x,
            self.cluster_centers_,
            dim="mean",
            metric=metric,
            metric_params=metric_params,
        )


class _KMedoidsCluster:
    def __init__(self, dist, cluster_idx):
        self._dist = dist
        self._cluster_idx = cluster_idx
        self._prev_cluster_idx = cluster_idx.copy()

    def assign(self):
        self._prev_cluster_idx[:] = self._cluster_idx
        self.labels_ = self._dist[self._cluster_idx].argmin(axis=0)

    def cost(self):
        self.cost_ = (
            np.take(self._dist, self._cluster_idx[self.labels_]).sum()
            / self._dist.shape[0]
        )  # np.sum(self._prev_cluster_idx != self._cluster_idx)

    def update(self):
        pass


class _FastKMedoidsCluster(_KMedoidsCluster):
    def update(self):
        for idx in range(self._cluster_idx.shape[0]):
            cluster_idx = np.where(self.labels_ == idx)[0]
            if cluster_idx.shape[0] == 0:
                continue

            cost = self._dist[cluster_idx, cluster_idx.reshape(-1, 1)].sum(axis=1)
            min_idx = cost.argmin()
            min_cost = cost[min_idx]
            curr_cost = cost[(cluster_idx == self._cluster_idx[idx]).argmax()]
            if min_cost < curr_cost:
                self._cluster_idx[idx] = cluster_idx[min_idx]


class _PamKMedoidsCluster(_KMedoidsCluster):
    def __init__(self, dist, cluster_idx):
        super().__init__(dist, cluster_idx)
        self._djs, self._ejs = np.sort(self._dist[self._cluster_idx], axis=0)[[0, 1]]

    def update(self):
        not_cluster_idx = np.delete(np.arange(self._dist.shape[0]), self._cluster_idx)
        optimal_swap = _pam_optimal_swap(
            self._dist,
            self._cluster_idx,
            not_cluster_idx,
            self._djs,
            self._ejs,
            self._cluster_idx.shape[0],
        )

        if optimal_swap is not None:
            i, j, _ = optimal_swap
            self._cluster_idx[self._cluster_idx == i] = j
            self._djs, self._ejs = np.sort(self._dist[self._cluster_idx], axis=0)[
                [0, 1]
            ]


_KMEDOID_CLUSTER_ALGORITHMS = {"pam": _PamKMedoidsCluster, "fast": _FastKMedoidsCluster}


class KMedoids(ClusterMixin, TransformerMixin, BaseEstimator):
    """
    KMedoid algorithm.

    Parameters
    ----------
    n_clusters : int, optional
        The number of clusters.
    metric : str, optional
        The metric.
    metric_params : dict, optional
        The metric parameters. Read more about the metrics and their parameters
        in the :ref:`User guide <list_of_metrics>`.
    init : {"auto", "random", "min"}, optional
        Cluster initialization. If "random", randomly initialize `n_clusters`,
        if "min" select the samples with the smallest distance to the other samples.
    n_init : "auto" or int, optional
        Number times the algorithm is re-initialized with new centroids.
    algorithm : {"fast", "pam"}, optional
        The algorithm for updating cluster assignments. If "pam", use the
        Partitioning Around Medoids algorithm.
    max_iter : int, optional
        The maximum number of iterations for a single run of the algorithm.
    tol : float, optional
        Relative tolerance to declare convergence of two consecutive iterations.
    verbose : int, optional
        Print diagnostic messages during convergence.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of `None` means using
        a single core and a value of `-1` means using all cores. Positive
        integers mean the exact number of cores.
    random_state : RandomState or int, optional
        Determines random number generation for centroid initialization and
        barycentering when fitting with `metric="dtw"`.

    Attributes
    ----------
    n_iter_ : int
        The number of iterations before convergence.
    cluster_centers_ : ndarray of shape (n_clusters, n_timestep)
        The cluster centers.
    medoid_indices_ : ndarray of shape (n_clusters, )
        The index of the medoid in the input samples.
    labels_ : ndarray of shape (n_samples, )
        The cluster assignment.
    """

    _parameter_constraints: dict = {
        "n_clusters": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [StrOptions(_METRICS.keys() | {"precomputed"})],
        "metric_params": [None, dict],
        "init": [StrOptions({"random", "auto", "min"})],
        "n_init": [
            StrOptions({"auto"}),
            Interval(numbers.Integral, 1, None, closed="left"),
        ],
        "algorithm": [StrOptions({"fast", "pam"})],
        "max_iter": [Interval(numbers.Integral, 1, None, closed="left")],
        "tol": [Interval(numbers.Real, 0, None, closed="left")],
        "verbose": [numbers.Integral],
        "n_jobs": [None, numbers.Integral],
        "random_state": ["random_state", None],
    }

    def __init__(
        self,
        n_clusters=8,
        metric="euclidean",
        metric_params=None,
        init="random",
        n_init="auto",
        algorithm="fast",
        max_iter=30,
        tol=1e-4,
        verbose=0,
        n_jobs=None,
        random_state=None,
    ):
        self.n_clusters = n_clusters
        self.metric = metric
        self.metric_params = metric_params
        self.init = init
        self.n_init = n_init
        self.algorithm = algorithm
        self.max_iter = max_iter
        self.tol = tol
        self.verbose = verbose
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, x, y=None):
        """
        Compute the kmedoids-clustering.

        Parameters
        ----------
        x : univariate time-series
            The input samples.
        y : Ignored, optional
            Not used.

        Returns
        -------
        object
            Fitted estimator.
        """
        x = self._validate_data(x, allow_3d=True)
        self._validate_params()

        if self.n_init == "auto":
            if self.init == "auto":
                n_init = 1
            else:
                n_init = 10
        else:
            n_init = self.n_init

        # initial medoids are deterministic
        if n_init > 1 and self.init != "random":
            n_init = 1

        max_iter = self.max_iter
        if self.algorithm == "pam" and self.n_clusters == 1 and self.max_iter != 0:
            warnings.warn("n_clusters must be larger than 1 if max_iter larger than 0")
            max_iter = 0

        random_state = check_random_state(self.random_state)
        if self.metric == "precomputed":
            dist = x
        else:
            dist = pairwise_distance(
                x,
                dim="mean",
                metric=self.metric,
                metric_params=self.metric_params,
                n_jobs=self.n_jobs,
            )
        best_iter = 0
        best_cost = np.inf
        best_clusterer = None
        best_reassign = None

        # TODO: Add sampling
        for i in range(n_init):
            if self.verbose:
                print(f"Running initialization {i}/{n_init}")

            iter, clusterer, reassign = self._fit_one_init(
                dist, random_state=random_state, max_iter=max_iter
            )

            if clusterer.cost_ < best_cost:
                best_cost = clusterer.cost_
                best_clusterer = clusterer
                best_iter = iter
                best_reassign = reassign

        if best_reassign:
            clusterer.assign()

        if self.metric == "precomputed":
            self.cluster_centers_ = None
        else:
            self.cluster_centers_ = x[best_clusterer._cluster_idx]

        self.inertia_ = best_clusterer.cost_
        self.medoid_indices_ = best_clusterer._cluster_idx
        self.n_iter_ = best_iter
        self.labels_ = best_clusterer.labels_
        return self

    def _fit_one_init(self, dist, *, max_iter, random_state):
        centers = self._init_centers(dist, self.n_clusters, random_state=random_state)
        clusterer = _KMEDOID_CLUSTER_ALGORITHMS[self.algorithm](dist, centers)
        clusterer.assign()
        reassign = False
        prev_cost = np.inf
        for iter in range(max_iter):
            clusterer.update()
            clusterer.cost()

            if self.verbose:
                print(
                    f"Iteration {iter}/{self.max_iter}: current_cost={clusterer.cost_}"
                )

            if math.isclose(clusterer.cost_, prev_cost, rel_tol=self.tol):
                reassign = True
                break

            prev_cost = clusterer.cost_
            clusterer.assign()

        if iter + 1 == self.max_iter:
            warnings.warn(
                "Maximum number of iterations reached before convergence. "
                "Consider increasing max_iter to improve the fit",
                ConvergenceWarning,
            )

        return iter, clusterer, reassign

    def _init_centers(self, dist, n_clusters, random_state):
        if self.init == "random":
            return random_state.choice(dist.shape[0], n_clusters, replace=False)
        elif self.init == "min" or (self.algorithm == "fast" and self.init == "auto"):
            return np.argpartition(dist.sum(axis=1), n_clusters)[:n_clusters]
        elif self.init == "auto":
            return _pam_build(dist, n_clusters)
        else:
            raise ValueError(f"Unsupported init, got {self.init}")

    def predict(self, x):
        """
        Predict the closest cluster for each sample.

        Parameters
        ----------
        x : univariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, )
            Index of the cluster each sample belongs to.
        """
        return self.transform(x).argmin(axis=1)

    def transform(self, x):
        """
        Transform the input to a cluster distance space.

        Parameters
        ----------
        x : univariate time-series
            The input samples.

        Returns
        -------
        ndarray of shape (n_samples, n_clusters)
            The distance between each sample and each cluster.
        """
        check_is_fitted(self)
        x = self._validate_data(x, allow_3d=True, reset=False)
        if self.metric == "precomputed":
            return x[:, self.medoid_indices_]
        else:
            return pairwise_distance(
                x,
                self.cluster_centers_,
                dim="mean",
                metric=self.metric,
                metric_params=self.metric_params,
            )



# Authors: Isak Samsten
# License: BSD 3 clause
"""Transform raw time series to tabular representations."""

from ._conv import convolve
from ._diff import DerivativeTransform, DiffTransform
from ._hydra import HydraTransform
from ._interval import FeatureTransform, IntervalTransform
from ._matrix_profile import MatrixProfileTransform
from ._pivot import PivotTransform, ProximityTransform
from ._rocket import RocketTransform
from ._sax import (
    PAA,
    SAX,
    piecewice_aggregate_approximation,
    symbolic_aggregate_approximation,
)
from ._shapelet import (
    CastorTransform,
    DilatedShapeletTransform,
    RandomShapeletTransform,
)

__all__ = [
    "convolve",
    "symbolic_aggregate_approximation",
    "piecewice_aggregate_approximation",
    "RandomShapeletTransform",
    "RocketTransform",
    "IntervalTransform",
    "PivotTransform",
    "ProximityTransform",
    "FeatureTransform",
    "DilatedShapeletTransform",
    "SAX",
    "PAA",
    "MatrixProfileTransform",
    "HydraTransform",
    "DiffTransform",
    "DerivativeTransform",
    "CastorTransform",
]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import os
import pkgutil
import pathlib
import sys

from importlib import metadata
from sphinx.util.logging import getLogger

import wildboar

sys.path.insert(0, os.path.abspath("_gen_figures"))
sys.path.insert(0, os.path.abspath(".") + "/_extensions")

logger = getLogger(__name__)

release = wildboar.__version__
version = release

if ".dev" in version:
    version_match = "dev"
else:
    version_match = ".".join(version.split(".")[:2])

logger.info("Building documentation for %s, matching %s", version, version_match)
# -- Project information -----------------------------------------------------

project = "Wildboar"
description = "Time series learning with Python"
copyright = "2023, Isak Samsten"
author = "Isak Samsten"

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "autoapi.extension",
    "lightdarkimg",
    "matplotlib.sphinxext.plot_directive",
    "nbsphinx",
    "numpydoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_inline_tabs",
    "sphinx_toggleprompt",
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}

pygments_style = "github-light-colorblind"
pygments_dark_style = "github-dark-colorblind"
syntax_highlight = "short"
add_function_parentheses = False

autoapi_dirs = ["../src/"]
autoapi_root = "api"
autoapi_template_dir = "_templates/autoapi/"
autoapi_ignore = ["*tests*", "_*.py"]
autoapi_keep_files = True
autoapi_add_toctree_entry = True
autoapi_python_class_content = "class"
autoapi_member_order = "groupwise"

autoapi_options = [
    "members",
    # "undoc-members",
    # "show-inheritance",
    "show-module-summary",
    # "special-members",
    "imported-members",
    "inherited-members",
]

html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "logo": {
        "text": "Wildboar",
        "image_light": "logo.png",
        "image_dark": "logo.png",
    },
    "show_version_warning_banner": True,
    "navbar_start": ["navbar-logo", "version-switcher"],
    "switcher": {
        "json_url": "https://wildboar.dev/_versions.json",
        "version_match": version_match,
    },
    "check_switcher": False,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/wildboar-foundation/wildboar/",
            "icon": "fa-brands fa-github",
            "type": "fontawesome",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/wildboar/",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "navigation_with_keys": False,
}

html_context = {
    "default_mode": "auto",
}

html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
    "css/pygments-override.css",
]
html_js_files = ["js/fa.js"]

rst_prolog = """
.. role:: python(code)
   :language: python
   :class: highlight
"""


# Find the source file given a module
def find_source(info):
    import importlib
    import inspect

    obj = importlib.import_module(info["module"])
    for part in info["fullname"].split("."):
        obj = getattr(obj, part)

    fn = os.path.normpath(inspect.getsourcefile(obj))
    fn_split = fn.split(os.sep)
    fn_index = fn_split.index("wildboar")
    fn = os.path.join(*fn_split[fn_index:])
    source, lineno = inspect.getsourcelines(obj)
    return fn, lineno, lineno + len(source) - 1


def linkcode_resolve(domain, info):
    if domain != "py" or not info["module"]:
        return None
    try:
        filename = "%s#L%d-L%d" % find_source(info)
    except:
        filename = info["module"].replace(".", "/") + ".py"

    return "https://github.com/wildboar-foundation/wildboar/blob/%s/src/%s" % (
        "master" if ".dev" in version else "v%s" % version,
        filename,
    )


regenerate_plots = os.environ.get("REGENERATE_PLOTS", False)

basepath = pathlib.Path("_static", "fig")
for modinfo in pkgutil.iter_modules(["_gen_figures"]):
    if modinfo.name.startswith("gen"):
        logger.info(f"Generating figures for: '{modinfo.name}'")
        module = modinfo.module_finder.find_module(modinfo.name).load_module(
            modinfo.name
        )
        funcs = list(module.__dict__.values())
        for func in funcs:
            if callable(func) and func.__name__.startswith("gen_"):
                logger.info(f" - '{func.__name__}'")
                func(basepath, force=regenerate_plots)



# cython: language_level=3

# Authors: Isak Samsten
# License: BSD 3 clause
from numpy cimport uint32_t


cdef enum:
    RAND_R_MAX = 2147483647

cdef struct VoseRand:
    Py_ssize_t n
    Py_ssize_t *alias
    double *prob

cdef void vose_rand_init(VoseRand *vr, Py_ssize_t n) noexcept nogil

cdef void vose_rand_free(VoseRand *vr) noexcept nogil

cdef void vose_rand_precompute(VoseRand *vr, const double *p) noexcept nogil

cdef Py_ssize_t vose_rand_int(VoseRand *vr, uint32_t *seed) noexcept nogil

cdef uint32_t rand_r(uint32_t *seed) noexcept nogil

cdef int rand_int(int min_val, int max_val, uint32_t *seed) noexcept nogil

cdef double rand_uniform(double low, double high, uint32_t *random_state) noexcept nogil

cdef double rand_normal(double mean, double std, uint32_t *random_state) noexcept nogil

cdef void shuffle(Py_ssize_t *values, Py_ssize_t length, uint32_t *seed) noexcept nogil

cdef class RandomSampler:

    cdef const double[::1] weights

    cdef Py_ssize_t upper

    cdef VoseRand vr

    cdef Py_ssize_t rand_int(self, uint32_t *seed) noexcept nogil



import itertools
import re
from collections import defaultdict

import numpy as np

from ..utils.validation import check_option
from ._distance import _METRICS, _SUBSEQUENCE_METRICS


def parse_metric_spec(kwargs):
    specs = defaultdict(dict)
    for key, value in kwargs.items():
        m = re.match(r"^(num|max|min)_([a-zA-Z_$]\w*)$", key)
        if m:
            specs[m.group(2)][m.group(1)] = value
        else:
            raise ValueError(
                f"The parameter {key} must be prefixed with "
                f"'min_', 'max_' or 'num_', got {key} "
            )

    return specs


def make_parameter_grid(metric_spec, default_n=10):
    if metric_spec is None:
        return [{}]

    specs = parse_metric_spec(metric_spec)
    params = []
    grids = []
    for param, spec in specs.items():
        if "max" not in spec:
            raise ValueError(f"The maximum value is missing for {param}.")

        if "min" not in spec:
            raise ValueError(f"The minimum value is missing for {param}.")

        if "num" in spec:
            num = spec["num"]
        else:
            num = default_n

        params.append(param)
        grids.append(np.linspace(spec["min"], spec["max"], num))

    return [
        {param: value for param, value in zip(params, grid)}
        for grid in itertools.product(*grids)
    ]


def _make_metric(metrics, metric, default_n=10, **kwargs):
    Metric = check_option(metrics, metric, "metric")
    if kwargs is None or default_n is None:
        return [Metric()]
    else:
        parameter_grid = make_parameter_grid(kwargs, default_n=default_n)
        return [Metric(**metric_params) for metric_params in parameter_grid]


def make_subsequence_metric(metric, *, default_n=10, **kwargs):
    return _make_metric(_SUBSEQUENCE_METRICS, metric, default_n=default_n, **kwargs)


def make_metric(metric, *, default_n=10, **kwargs):
    return _make_metric(_METRICS, metric, default_n=default_n, **kwargs)


def _make_metrics(metric_specs, factory):
    metrics = []
    weights = []
    base_weight = 1.0 / len(metric_specs)

    if isinstance(metric_specs, dict):
        metric_specs = metric_specs.items()
    for metric_name, metric_spec in metric_specs:
        metric_spec = {} if metric_spec is None else metric_spec
        current_metrics = factory(metric_name, **metric_spec)
        for metric in current_metrics:
            metrics.append(metric)
            weights.append(base_weight / len(current_metrics))

    return metrics, np.array(weights, dtype=float)


def make_metrics(metric_specs):
    return _make_metrics(metric_specs, make_metric)


def make_subsequence_metrics(metric_specs):
    return _make_metrics(metric_specs, make_subsequence_metric)



# Authors: Isak Samsten
# License: BSD 3 clause

from ..transform import RocketTransform
from ._transform import TransformRidgeClassifierCV, TransformRidgeCV


class RocketClassifier(TransformRidgeClassifierCV):
    """Implements the ROCKET classifier."""

    def __init__(  # noqa: PLR0913
        self,
        n_kernels=10000,
        *,
        kernel_size=None,
        sampling="normal",
        sampling_params=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        class_weight=None,
        normalize=True,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            scoring=scoring,
            cv=cv,
            class_weight=class_weight,
            n_jobs=n_jobs,
            normalize=normalize,
            random_state=random_state,
        )
        self.padding_prob = padding_prob
        self.normalize_prob = normalize_prob
        self.bias_prob = bias_prob
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.n_kernels = n_kernels

    def _get_transform(self, random_state):
        return RocketTransform(
            self.n_kernels,
            kernel_size=self.kernel_size,
            sampling=self.sampling,
            sampling_params=self.sampling_params,
            bias_prob=self.bias_prob,
            normalize_prob=self.normalize_prob,
            padding_prob=self.padding_prob,
            random_state=random_state,
            n_jobs=self.n_jobs,
        )


class RocketRegressor(TransformRidgeCV):
    """Implements the ROCKET regressor."""

    def __init__(
        self,
        n_kernels=10000,
        *,
        alphas=(0.1, 1.0, 10.0),
        fit_intercept=True,
        scoring=None,
        cv=None,
        gcv_mode=None,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(
            alphas=alphas,
            fit_intercept=fit_intercept,
            scoring=scoring,
            cv=cv,
            gcv_mode=gcv_mode,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.n_kernels = n_kernels

    def _get_transform(self, random_state):
        return RocketTransform(
            self.n_kernels, random_state=random_state, n_jobs=self.n_jobs
        )



# Authors: Isak Samsten
# License: BSD 3 clause
import numbers

import numpy as np
from sklearn.base import TransformerMixin
from sklearn.utils import check_random_state
from sklearn.utils._param_validation import Interval, StrOptions

from ..base import BaseEstimator
from ..distance._distance import _METRICS
from ..distance._multi_metric import make_metrics
from ..utils._rand import RandomSampler
from ..utils.validation import check_option
from ._base import BaseAttributeTransform
from ._cpivot import PivotAttributeGenerator

_METRIC_NAMES = set(_METRICS.keys())
_METRIC_NAMES.add("auto")
_METRIC_NAMES = frozenset(_METRIC_NAMES)


class PivotMixin:
    """Mixin for Pivot based estimators."""

    _parameter_constraints: dict = {
        "n_pivots": [Interval(numbers.Integral, 1, None, closed="left")],
        "metric": [StrOptions(_METRIC_NAMES), list],
        "metric_params": [None, dict],
        "metric_sample": [StrOptions({"uniform", "weighted"}), None],
    }

    def _get_generator(self, x, y):
        if isinstance(self.metric, str) and self.metric == "auto":
            metric_specs = [
                ("euclidean", None),
                ("dtw", None),
                ("ddtw", None),
                ("dtw", dict(min_r=0.0, max_r=0.25, num_r=10)),
                ("ddtw", dict(min_r=0.0, max_r=0.25, num_r=10)),
                ("wdtw", dict(min_g=0.2, max_g=1.0)),
                ("wddtw", dict(min_g=0.2, max_g=1.0)),
                (
                    "lcss",
                    dict(
                        min_r=0.0,
                        max_r=0.25,
                        min_epsilon=0.2,
                        max_epsilon=1.0,
                    ),
                ),
                ("erp", dict(min_g=0, max_g=1.0)),
                ("msm", dict(min_c=0.01, max_c=100, num_c=50)),
                (
                    "twe",
                    dict(
                        min_penalty=0.00001,
                        max_penalty=1.0,
                        min_stiffness=0.000001,
                        max_stiffness=0.1,
                    ),
                ),
            ]
            metrics, weights = make_metrics(metric_specs)
        elif isinstance(self.metric, str):
            Metric = check_option(_METRICS, self.metric, "metric")
            metric_params = self.metric_params if self.metric_params is not None else {}
            metrics = [Metric(**metric_params)]
            weights = np.ones(1)
        else:
            metrics, weights = make_metrics(self.metric)

        if self.metric_sample is None or self.metric_sample == "uniform":
            weights = None

        return PivotAttributeGenerator(
            self.n_pivots, metrics, RandomSampler(len(metrics), weights)
        )


class PivotTransform(PivotMixin, BaseAttributeTransform):
    """
    A transform using pivot time series and sampled distance metrics.

    Parameters
    ----------
    n_pivots : int, optional
        The number of pivot time series.
    metric : {'auto'} or list, optional
        - If str, the metric to compute the distance.
        - If list, multiple metrics specified as a list of tuples, where the first
            element of the tuple is a metric name and the second element a dictionary
            with a parameter grid specification. A parameter grid specification is a
            dict with two mandatory and one optional key-value pairs defining the
            lower and upper bound on the values and number of values in the grid. For
            example, to specifiy a grid over the argument 'r' with 10 values in the
            range 0 to 1, we would give the following specification: `dict(min_r=0,
            max_r=1, num_r=10)`.

        Read more about the metrics and their parameters in the :ref:`User guide
        <list_of_subsequence_metrics>`.
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the :ref:`User guide <list_of_metrics>`.
    metric_sample : {"uniform", "weighted"}, optional
        If multiple metrics are specified this parameter controls how they are
        sampled. "uniform" samples each metric configuration with equal probability
        and "weighted" samples each metric with equal probability. By default,
        metric configurations are sampled with equal probability.
    random_state : int or np.RandomState, optional
        The random state.
    n_jobs : int, optional
        The number of cores to use.
    """

    _parameter_constraints: dict = {
        **PivotMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_pivots=100,
        *,
        metric="auto",
        metric_params=None,
        metric_sample=None,
        random_state=None,
        n_jobs=None,
    ):
        super().__init__(random_state=random_state, n_jobs=n_jobs)
        self.n_pivots = n_pivots
        self.metric = metric
        self.metric_params = metric_params
        self.metric_sample = metric_sample


class ProximityTransform(TransformerMixin, BaseEstimator):
    """
    Transform time series based on class conditional pivots.

    Parameters
    ----------
    n_pivots : int, optional
        The number of pivot time series per class.
    metric : {'auto'} or list, optional
        - If str, the metric to compute the distance.
        - If list, multiple metrics specified as a list of tuples, where the first
            element of the tuple is a metric name and the second element a dictionary
            with a parameter grid specification. A parameter grid specification is a
            dict with two mandatory and one optional key-value pairs defining the
            lower and upper bound on the values and number of values in the grid. For
            example, to specifiy a grid over the argument 'r' with 10 values in the
            range 0 to 1, we would give the following specification: `dict(min_r=0,
            max_r=1, num_r=10)`.

        Read more about the metrics and their parameters in the :ref:`User guide
        <list_of_subsequence_metrics>`.
    metric_params : dict, optional
        Parameters for the distance measure. Ignored unless metric is a string.

        Read more about the parameters in the :ref:`User guide <list_of_metrics>`.
    metric_sample : {"uniform", "weighted"}, optional
        If multiple metrics are specified this parameter controls how they are
        sampled. "uniform" samples each metric configuration with equal probability
        and "weighted" samples each metric with equal probability. By default,
        metric configurations are sampled with equal probability.
    random_state : int or np.RandomState, optional
        The random state.
    n_jobs : int, optional
        The number of cores to use.
    """

    def __init__(
        self,
        n_pivots=100,
        metric="auto",
        metric_params=None,
        metric_sample="weighted",
        random_state=None,
        n_jobs=None,
    ):
        self.n_pivots = n_pivots
        self.metric = metric
        self.metric_params = metric_params
        self.metric_sample = metric_sample
        self.random_state = random_state
        self.n_jobs = n_jobs

    def fit(self, X, y):
        X, y = self._validate_data(X, y, allow_3d=True, dtype=float)
        random_state = check_random_state(self.random_state)
        self.pivots_ = [
            PivotTransform(
                n_pivots=int(min(self.n_pivots, count)),
                metric=self.metric,
                metric_params=self.metric_params,
                metric_sample=self.metric_sample,
                random_state=random_state.randint(np.iinfo(np.int32).max),
                n_jobs=self.n_jobs,
            ).fit(X[y == label])
            for label, count in zip(*np.unique(y, return_counts=True))
        ]

        return self

    def transform(self, X, y=None):
        return np.hstack([pivots.transform(X) for pivots in self.pivots_])



# cython: language_level=3
# cython: boundscheck=False
# cython: cdivision=True
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, sqrt
from libc.stdlib cimport free, malloc

from ..utils cimport TSArray
from ..utils._rand cimport RAND_R_MAX
from ..utils._stats cimport (
    IncStats,
    cumulative_mean_std,
    find_min,
    inc_stats_add,
    inc_stats_init,
    inc_stats_remove,
    inc_stats_variance,
)
from ._mass cimport _mass_distance


cdef double EPSILON = 1e-13

cdef void _matrix_profile_stmp(
    const double *x,
    Py_ssize_t x_length,
    const double *y,
    Py_ssize_t y_length,
    Py_ssize_t window,
    Py_ssize_t exclude,
    double *mean_x,
    double *std_x,
    complex *x_buffer,
    complex *y_buffer,
    double *dist_buffer,
    double *mp,
    Py_ssize_t *mpi,
) noexcept nogil:
    """Compute the matrix profile using the STMP algorithm

    Parameters
    ----------
    x : double*
        The time series.
    x_length : int
        The size of x.
    y : double*
        A time series.
    y_length : int
        The size of y
    window : int
        The size of subsequences used to form the profile.
    exclude : double
        The size of the exclusion zone, expressed as a fraction of `window`.
    mean_x : double*
        The buffer of cumulative mean of subsequences of size `window` in x, with size
        `profile_length`.
    std_x : double
        The buffer of cumulative std of subsequences of size `window` in x, with size
        `profile_length`.
    x_buffer : complex*
        The buffer used for the distance computation, with size `x_length`
    y_buffer : complex*
        The buffer used for the distance computation, with size `x_length`
    dist_buffer : double*
        The buffer used to store distances at the i:th iteration, with size
        `profile_length`.
    mp : double*
        The matrix profile, with size `profile_length`
    mpi : int*
        The matrix profile index, with size `profile_length`

    Notes
    =====
    - The `profile_length` is computed as `y_length - window + 1`
    - The buffer-parameters can be empty on entry.
    """
    cdef Py_ssize_t i, j
    cdef double std
    cdef IncStats stats
    cdef Py_ssize_t profile_length = y_length - window + 1
    cumulative_mean_std(x, x_length, window, mean_x, std_x)
    inc_stats_init(&stats)
    for i in range(window - 1):
        inc_stats_add(&stats, 1.0, y[i])

    for i in range(profile_length):
        mp[i] = INFINITY
        mpi[i] = -1

    for i in range(profile_length):
        inc_stats_add(&stats, 1.0, y[i + window - 1])
        std = sqrt(inc_stats_variance(&stats))
        _mass_distance(
            x,
            x_length,
            y + i,
            window,
            stats.mean,
            std,
            mean_x,
            std_x,
            x_buffer,
            y_buffer,
            dist_buffer,
        )
        inc_stats_remove(&stats, 1.0, y[i])
        for j in range(x_length - window + 1):
            if dist_buffer[j] < mp[i] and (j <= i - exclude or j >= i + exclude):
                mp[i] = dist_buffer[j]
                mpi[i] = j


def _paired_matrix_profile(
    TSArray X,
    TSArray Y,
    Py_ssize_t w,
    Py_ssize_t dim,
    Py_ssize_t exclude,
    n_jobs,
):
    cdef Py_ssize_t profile_length = Y.shape[2] - w + 1
    cdef Py_ssize_t i
    cdef double *mean_x = <double*> malloc(sizeof(double) * X.shape[2])
    cdef double *std_x = <double*> malloc(sizeof(double) * X.shape[2])
    cdef double *dist_buffer = <double*> malloc(sizeof(double) * X.shape[2])
    cdef complex *x_buffer = <complex*> malloc(sizeof(complex) * X.shape[2])
    cdef complex *y_buffer = <complex*> malloc(sizeof(complex) * X.shape[2])

    cdef double[:, :] mp = np.empty((X.shape[0], profile_length), dtype=np.double)
    cdef Py_ssize_t[:, :] mpi = np.empty((X.shape[0], profile_length), dtype=np.intp)

    with nogil:
        for i in range(X.shape[0]):
            _matrix_profile_stmp(
                &X[i, dim, 0],
                X.shape[2],
                &Y[i, dim, 0],
                Y.shape[2],
                w,
                exclude,
                mean_x,
                std_x,
                x_buffer,
                y_buffer,
                dist_buffer,
                &mp[i, 0],
                &mpi[i, 0],
            )
    free(mean_x)
    free(std_x)
    free(dist_buffer)
    free(x_buffer)
    free(y_buffer)
    return mp.base, mpi.base



# Authors: Isak Samsten
# License: BSD 3 clause

import base64
import hashlib
import os
import pickle
import re
import sys
import warnings
import zipfile
from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.utils.fixes import parse_version

from .. import EOS
from .. import __version__ as wildboar_version

try:
    import requests
except ModuleNotFoundError as e:
    from ..utils import _soft_dependency_error

    _soft_dependency_error(e, context="wildboar.datasets")

DEFAULT_TAG = "default"


def _replace_placeholders(url, **kwargs):
    """
    Replace placeholder values of the format {key} with kwargs[key].

    Parameters
    ----------
    url : str
        The input string
    **kwargs : dict
        The key and values to replace

    Returns
    -------
    str : url with placeholder values replaced
    """
    for arg, value in kwargs.items():
        url = url.replace("{%s}" % arg, value)
    return url


def _check_integrity(bundle_file, hash_file, throws=True):
    """
    Check the integrity of the downloaded or cached file.

    Parameters
    ----------
    bundle_file : str, bytes or PathLike
        Path to the bundle file
    hash_file : str, bytes or PathLike
        Path to the hash file
    throws : bool
        Throw an exception on hash missmatch

    Returns
    -------
    bool
        true if the hash of bundle file matches the contents of the hash file
    """
    with open(hash_file, "r") as f:
        hash = f.readline().strip()

    if hash is not None:
        actual_hash = _sha1(bundle_file)
        if hash != actual_hash:
            if throws:
                raise ValueError(
                    "Integrity check failed, expected '%s', got '%s'."
                    % (hash, actual_hash)
                )
            else:
                return False
    return True


def _sha1_is_sane(hash_file):
    """
    Check the sanity of a hash file.

    Parameters
    ----------
    hash_file : str, bytes or PathLike
        The path to the hash file

    Returns
    -------
    bool
        Returns true if hash is 40 characters long; otherwise false.
    """
    with open(hash_file, "r") as f:
        return len(f.readline().strip()) == 40


def _load_archive(
    bundle_name,
    download_url,
    cache_dir,
    *,
    create_cache_dir=True,
    progress=True,
    force=False,
):
    """
    Load or download a bundle.

    Parameters
    ----------
    bundle_name : str
        The name of the cached file.
    download_url : str
        The download url to the bundle and hash file.
    cache_dir : str
        The cache directory.
    create_cache_dir : bool
        Create the cache directory if missing.
    progress : bool
        Show progress information
    force : bool
        Remove any cached files and force re-download.

    Returns
    -------
    zipfile.ZipFile
        A zip-archive with datasets.
    """
    if not os.path.exists(cache_dir):
        if create_cache_dir:
            os.makedirs(os.path.abspath(cache_dir), exist_ok=True)
        else:
            raise ValueError(
                "The output directory %s does not exist. "
                "Set create_cache_dir=True to create it." % cache_dir
            )

    cached_hash = os.path.join(cache_dir, "%s.sha1" % bundle_name)
    cached_bundle = os.path.join(cache_dir, "%s.zip" % bundle_name)

    if force:
        if os.path.exists(cached_hash):
            os.remove(cached_hash)
        if os.path.exists(cached_bundle):
            os.remove(cached_bundle)

    if not os.path.exists(cached_hash) or not _sha1_is_sane(cached_hash):
        _download_hash_file(cached_hash, "%s.sha1" % download_url, bundle_name)

    if os.path.exists(cached_bundle) and _check_integrity(
        cached_bundle, cached_hash, throws=False
    ):
        try:
            return zipfile.ZipFile(open(cached_bundle, "rb"))
        except zipfile.BadZipFile:
            os.remove(cached_bundle)

    _download_bundle_file(cached_bundle, "%s.zip" % download_url, bundle_name, progress)
    _check_integrity(cached_bundle, cached_hash)
    return zipfile.ZipFile(open(cached_bundle, "rb"))


def _download_hash_file(cached_hash, hash_url, filename):
    """
    Download the hash file.

    Parameters
    ----------
    cached_hash : str, bytes or PathLike
        The path to the cached hash.
    hash_url : str
        The download url.
    filename : str
        The filename of the bundle.
    """
    with open(cached_hash, "w") as f:
        response = requests.get(hash_url)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            f.close()
            os.remove(cached_hash)
            raise ValueError(
                "Downloading %s.sha1 failed. Try another version or tag." % filename
            ) from e

        f.write(response.text)


_SIZE_NAMES = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")


def _convert_size(byte_size, i=None):
    import math

    if byte_size == 0:
        return 0, 0

    if i is None:
        i = int(math.floor(math.log(byte_size, 1024)))
    p = math.pow(1024, i)
    s = round(byte_size / p, 2)
    return s, i


def _print_progress(
    current_iter,
    max_iter,
    *,
    prefix="",
    suffix="",
    progress=None,
    length=50,
    fill="█",
    unfill="-",
    end="\r",
    file=sys.stderr,
):
    def size_progress(current, total):
        total_size, total_type = _convert_size(total)
        current_size, _ = _convert_size(current, total_type)
        return f" {current_size}/{total_size} {_SIZE_NAMES[total_type]}"

    if progress is None:
        progress = size_progress

    progress = progress(float(current_iter), float(max_iter))
    fill_size = int(length * current_iter // max_iter)
    unfill_size = int(length - fill_size)
    bar = fill * fill_size + unfill * unfill_size
    print(f"\r{prefix}|{bar}|{progress}{suffix}", file=file, end=end)
    if current_iter >= max_iter:
        print(file=file)


def _download_bundle_file(cached_bundle, bundle_url, filename, progress):
    """
    Download the bundle.

    Parameters
    ----------
    cached_bundle : str, bytes or PathLike
        The path to the cached bundle
    bundle_url : str
        The download url
    filename : str
        The filename of the bundle
    progress : bool
        Show progress bar
    """
    with open(cached_bundle, "wb") as f:
        response = requests.get(bundle_url, stream=True)
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            f.close()
            os.remove(cached_bundle)
            raise ValueError(
                "Downloading %s.zip failed. Try another version or tag." % filename
            ) from e

        total_length = response.headers.get("content-length")
        if total_length is None:  # no content length header
            f.write(response.content)
        else:
            length = 0
            total_length = int(total_length)

            if progress:
                total_size, total_type = _convert_size(total_length)
                print(
                    f"Downloading {filename}.zip "
                    f"({total_size} {_SIZE_NAMES[total_type]})",
                    file=sys.stderr,
                )

            for data in response.iter_content(chunk_size=4096):
                length += len(data)
                f.write(data)
                if progress:
                    _print_progress(
                        length,
                        total_length,
                        prefix="  ",
                        unfill="-",
                    )


class Repository(metaclass=ABCMeta):
    """A repository is a collection of bundles."""

    def __init__(self):
        self._active = False

    def __eq__(self, o):
        if isinstance(o, Repository):
            return self.identifier == o.identifier
        return False

    def __hash__(self) -> int:
        return hash(self.identifier)

    @property
    @abstractmethod
    def identifier(self):
        pass

    @property
    @abstractmethod
    def name(self):
        """
        Name of the repository.

        Returns
        -------
        str
            The name of the repository.
        """
        pass

    @property
    @abstractmethod
    def version(self):
        """
        The repository version.

        Returns
        -------
        str
            The version of the repository.
        """
        pass

    @property
    @abstractmethod
    def download_url(self):
        """
        The url template for downloading bundles.

        Returns
        -------
        str
            The download url.
        """
        pass

    @property
    @abstractmethod
    def wildboar_requires(self):
        """
        The minimum required wildboar version.

        Returns
        -------
        str
            The min version.
        """
        pass

    @abstractmethod
    def get_bundles(self):
        """
        Get all bundles.

        Returns
        -------
        dict
            A dictionary of key and bundle.
        """
        pass

    @property
    def active(self):
        return self._active

    def get_bundle(self, key):
        """
        Get a bundle with the specified key.

        Parameters
        ----------
        key : str
            Key of the bundle.

        Returns
        -------
        Bundle, optional
            A bundle or None.
        """
        bundle = self.get_bundles().get(key)
        if bundle is None:
            raise ValueError("The bundle %s does not exist." % key)
        return bundle

    def load_dataset(
        self,
        bundle,
        dataset,
        *,
        cache_dir,
        tag=None,
        create_cache_dir=True,
        progress=True,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = bundle.version
        tag = tag or bundle.tag

        cache_dir = os.path.join(cache_dir, self.identifier)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )

        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.load(dataset, archive)

    def list_datasets(
        self,
        bundle,
        *,
        cache_dir,
        collection=None,
        tag=None,
        create_cache_dir=True,
        progress=True,
        force=False,
    ):
        bundle = self.get_bundle(bundle)
        version = bundle.version
        tag = tag or bundle.tag

        cache_dir = os.path.join(cache_dir, self.identifier)
        download_url = _replace_placeholders(
            self.download_url, bundle=bundle.key, version=version, tag=tag
        )
        with _load_archive(
            bundle_name=bundle.get_filename(version, tag),
            download_url=download_url,
            cache_dir=cache_dir,
            create_cache_dir=create_cache_dir,
            progress=progress,
            force=force,
        ) as archive:
            return bundle.list(archive, collection)

    def clear_cache(self, cache_dir, keep_last_version=True):
        cache_dir = os.path.join(cache_dir, self.identifier)
        if not os.path.exists(cache_dir) or not os.path.isdir(cache_dir):
            return

        keep = []
        if keep_last_version:
            keep = [
                bundle.get_filename(tag=bundle.tag)
                for _, bundle in self.get_bundles().items()
            ]

        for filename in os.listdir(cache_dir):
            basename, ext = os.path.splitext(filename)
            full_path = os.path.join(cache_dir, filename)
            if (
                os.path.isfile(full_path)
                and ext in [".zip", ".sha1"]
                and basename not in keep
            ):
                os.remove(full_path)

    def refresh(self, timeout=None):
        """Refresh the repository."""
        try:
            self._refresh(timeout)
            self._active = True
        except requests.ConnectionError:
            self._active = False

    @abstractmethod
    def _refresh(self, timeout):
        pass


def _validate_url(url):
    if "{bundle}" in url and "{version}" in url and "{tag}" in url:
        return url
    else:
        raise ValueError(
            "The repository url is invalid, got %s "
            "({bundle}, {version} and {tag} are required)" % url
        )


def _validate_repository_name(str):
    if re.match("[a-zA-Z]+", str):
        return str
    else:
        raise ValueError(
            "The repository name %s is not valid, expected only characters." % str
        )


def _validate_bundle_key(str):
    if re.match(r"[a-zA-Z0-9\-]+", str):
        return str
    else:
        raise ValueError(
            "The bundle key %s is not valid, expected only characters, numbers or -."
            % str
        )


def _validate_version(str, *, max_version=None):
    if re.match(r"(^(?:\d+\.)?(?:\d+\.)?(?:\*|\d+)$)", str):
        if max_version and parse_version(str) > parse_version(max_version):
            raise ValueError(
                "The dataset is not supported, required version >=%s, got %s"
                % (max_version, str)
            )
        return str
    else:
        raise ValueError("The version %s is not valid." % str)


def _validate_collections(collections):
    if not isinstance(collections, dict):
        raise TypeError(
            "'collections' must be dict, not %r" % type(collections).__qualname__
        )
    else:
        for key, values in collections.items():
            if not isinstance(values, list):
                raise TypeError(
                    "'collections.value' must be list, not %r"
                    % type(values).__qualname__
                )
            if not isinstance(key, str):
                raise TypeError(
                    "'collections.key' must be str, not" % type(key).__qualname__
                )
    return collections


class JSONRepository(Repository):
    supported_version = "1.1"

    def __init__(self, url):
        super().__init__()
        self.repo_url = url

    @property
    def identifier(self):
        return base64.b64encode(self.repo_url.encode()).decode("utf-8")

    @property
    def wildboar_requires(self):
        return self._wildboar_requires

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    @property
    def download_url(self):
        return self._bundle_url

    def get_bundles(self):
        return self._bundles

    def _refresh(self, timeout):
        json = requests.get(self.repo_url, timeout=timeout).json()
        self._wildboar_requires = json["wildboar_requires"]
        self._name = _validate_repository_name(json["name"])
        self._version = _validate_version(
            json["version"], max_version=JSONRepository.supported_version
        )
        if parse_version(self.wildboar_requires) > parse_version(wildboar_version):
            raise ValueError(
                "The repository requires wildboar >=%s, got %s"
                % (self.wildboar_requires, wildboar_version),
            )
        self._bundle_url = _validate_url(json["bundle_url"])
        bundles = {}
        for bundle_json in json["bundles"]:
            key = _validate_bundle_key(bundle_json["key"])
            if key in bundles:
                warnings.warn("duplicate dataset %s (ignoring)" % key, UserWarning)

            version = _validate_version(bundle_json["version"])
            tag = bundle_json.get("tag")
            if tag is not None:
                tag = _validate_bundle_key(tag)

            name = bundle_json.get("name")
            if name is None:
                raise ValueError("bundle %s require 'name', but got None" % key)

            arrays = bundle_json.get("arrays")
            if arrays is not None:
                if not isinstance(arrays, list):
                    warnings.warn(
                        "'arrays' should be list, not %r. Using defaults."
                        % type(arrays).__qualname__,
                        UserWarning,
                    )
                    arrays = None

            description = bundle_json.get("description")
            if description is not None:
                if not isinstance(description, str):
                    warnings.warn(
                        "'description' should be str, not %r. Using defaults."
                        % type(description).__qualname__,
                        UserWarning,
                    )
                    description = None

            collections = bundle_json.get("collections")
            if collections is not None:
                collections = _validate_collections(collections)

            bundles[bundle_json["key"]] = NpBundle(
                key=key,
                version=version,
                tag=tag,
                name=name,
                description=description,
                collections=collections,
                arrays=arrays,
            )

        self._bundles = bundles


class RepositoryCollection:
    def __init__(self):
        self.pending_repositories = set()
        self.repositories = set()
        self._initialized = False

    def __getitem__(self, key):
        if not self._initialized:
            self._initialized = True
            self.refresh(timeout=3)

        repository = next(
            (repository for repository in self.repositories if repository.name == key),
            None,
        )

        if repository is None:
            if self.pending_repositories:
                raise ValueError(
                    "The repository %s does not exist, but %d repositories have not "
                    "been refreshed yet." % (key, len(self.pending_repositories))
                )
            else:
                raise ValueError("The repository %s does not exist." % key)

        return repository

    def __delitem__(self, key):
        self.repositories = set(
            repository for repository in self.repositories if repository.name != key
        )

    def __contains__(self, item):
        return any(
            repository for repository in self.repositories if repository.name == item
        )

    def __iter__(self):
        return iter(self.repositories)

    def __len__(self):
        return len(self.repositories)

    def load_repository(self, repository, cache_dir=None):
        if cache_dir and not repository.active:
            cache_dir = os.path.join(cache_dir, repository.identifier)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)
            repo_path = os.path.join(cache_dir, "repo")
            if os.path.exists(repo_path):
                with open(repo_path, "rb") as f:
                    return pickle.load(f)

        return None

    def save_repository(self, repository, cache_dir=None):
        if cache_dir and repository.active:
            cache_dir = os.path.join(cache_dir, repository.identifier)
            if not os.path.exists(cache_dir):
                os.makedirs(cache_dir, exist_ok=True)
            repo_path = os.path.join(cache_dir, "repo")
            with open(repo_path, "wb") as f:
                pickle.dump(repository, f)

    def refresh(self, repository=None, timeout=None, cache_dir=None):
        if repository is None:
            for repository in self.repositories:
                repository.refresh(timeout)
                if repository.active:
                    self.save_repository(repository, cache_dir=cache_dir)

            for repository in self.pending_repositories:
                repository.refresh(timeout)
                if repository.active:
                    self._check_unique(repository.name)
                    self.repositories.add(repository)
                    self.save_repository(repository, cache_dir=cache_dir)
                else:
                    cached_repository = self.load_repository(
                        repository, cache_dir=cache_dir
                    )
                    if cached_repository:
                        self._check_unique(cached_repository.name)
                        self.repositories.add(cached_repository)

            self.pending_repositories = set(
                repository
                for repository in self.pending_repositories
                if not repository.active
            )
        else:
            repository = self[repository]
            repository.refresh(timeout)
            if repository.active:
                self.save_repository(repository, cache_dir=cache_dir)

    def install(self, repository, refresh=True, timeout=None, cache_dir=None):
        if refresh:
            repository.refresh(timeout)

        if repository.active:
            self._check_unique(repository.name)
            self.repositories.add(repository)
        else:
            cached_repository = self.load_repository(repository, cache_dir=cache_dir)
            if cached_repository:
                self._check_unique(cached_repository)
                self.repositories.add(cached_repository)
            else:
                self.pending_repositories.add(repository)

    def _check_unique(self, key):
        if key in {repo.name for repo in self.repositories}:
            raise ValueError(f"Repository with {key} already exists")


class Bundle(metaclass=ABCMeta):
    """
    Base class for handling dataset bundles.

    Parameters
    ----------
    key : str
        A unique key of the bundle.
    version : str
        The version of the bundle.
    name : str
        Human-readable name of the bundle.
    tag : str, optional
        A bundle tag.
    arrays : list
        The arrays of the dataset.
    description : str
        Description of the bundle.
    collections : dict, optional
        A list of collections.
    """

    def __init__(
        self,
        *,
        key,
        version,
        name,
        tag=None,
        arrays=None,
        description=None,
        collections=None,
    ):
        self.key = key
        self.version = version
        self.name = name
        self.description = description
        self.collections = collections
        self.tag = tag or DEFAULT_TAG
        self.arrays = arrays or ["x", "y"]

    def get_filename(self, version=None, tag=None, ext=None):
        """
        Get the cache name of the bundle.

        Parameters
        ----------
        version : str, optional
            The bundle version.
        tag : str, optional
            The tag.
        ext : str, optional
            The extension of the file.

        Returns
        -------
        str
            The filename.
        """
        filename = "%s-v%s" % (self.key, version or self.version)
        if tag:
            filename += "-%s" % tag
        if ext:
            filename += ext
        return filename

    def get_collection(self, collection):
        """
        Get a dataset collection.

        Parameters
        ----------
        collection : str, optional
            The name of the collection.

        Returns
        -------
        list
            List of datasets in the collection.
        """
        if self.collections is None:
            raise ValueError("The collection %s cannot be found." % collection)
        else:
            c = self.collections.get(collection)
            if c is None:
                raise ValueError("The collection %s cannot be found." % collection)
            return c

    def list(self, archive, collection=None):
        """
        List all datasets in this bundle.

        Parameters
        ----------
        archive : ZipFile
            The bundle file.
        collection : str, optional
            The collection name.

        Returns
        -------
        list
            A sorted list of datasets in the bundle.
        """
        names = []
        if collection is not None:
            sample = self.get_collection(collection)
        else:
            sample = None
        for f in archive.filelist:
            path, ext = os.path.splitext(f.filename)
            if self._is_dataset(path, ext):
                filename = os.path.basename(path)
                filename = re.sub("_(TRAIN|TEST)", "", filename)
                if sample is None or filename in sample:
                    names.append(filename)

        return sorted(set(names))

    def load(  # noqa: PLR0912
        self,
        name,
        archive,
    ):
        """
        Load a dataset from the bundle.

        Parameters
        ----------
        name : str
            Name of the dataset.
        archive : ZipFile
            The zip-file bundle.

        Returns
        -------
        x : ndarray
            Data samples.
        y : ndarray
            Data labels.
        n_training_samples : int
            Number of samples that are for training. The value is <= x.shape[0].
        extras : dict, optional
            Extra numpy arrays.
        """
        datasets = []
        for dataset in map(_Dataset, archive.filelist):
            if dataset.filename == name and self._is_dataset(dataset.path, dataset.ext):
                datasets.append(dataset)

        if not datasets:
            raise ValueError("The dataset %s cannot be found." % name)

        train_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "train"
        ]
        test_parts = [
            self._load_array(archive, dataset.file)
            for dataset in datasets
            if dataset.part == "test"
        ]

        data = {}
        for array in self.arrays:
            data[array] = np.concatenate(
                [train_part[array] for train_part in train_parts], axis=0
            )

        sizes = [data[array].shape[0] for array in self.arrays]
        if max(sizes) != min(sizes):
            raise ValueError(
                "Arrays %s must have the same number of samples." % self.arrays
            )

        n_train_samples = sizes[0]
        if test_parts:
            for array in self.arrays:
                train = data[array]
                test = []
                for test_part in test_parts:
                    test.append(test_part[array])
                test = np.concatenate(test, axis=0)

                if test.ndim != train.ndim:
                    raise ValueError(
                        "The training (%d) and testing (%d) parts have incompatible "
                        "rank" % (test.ndim, train.ndim)
                    )

                if test.ndim > 2 and train.shape[1] != test.shape[1]:
                    raise ValueError(
                        "The training (%d) and testing (%d) parts have incompatible "
                        "number dimensions" % (train.shape[1], test.shape[1])
                    )

                if test.ndim == 1:
                    data[array] = np.concatenate([train, test], axis=0)
                else:
                    if test.ndim == 2:
                        shape = (
                            train.shape[0] + test.shape[0],
                            max(test.shape[-1], train.shape[-1]),
                        )
                    elif test.ndim == 3:
                        shape = (
                            train.shape[0] + test.shape[0],
                            test.shape[1],
                            max(test.shape[-1], train.shape[-1]),
                        )
                    else:
                        raise ValueError(
                            "Invalid dataset rank %d, expected <= 3" % test.ndim
                        )

                    merge = np.full(shape, fill_value=EOS, dtype=np.double)
                    merge[: train.shape[0], ..., : train.shape[-1]] = train
                    merge[train.shape[0] :, ..., : test.shape[-1]] = test
                    data[array] = merge

        x = data.pop("x")
        y = data.pop("y")
        return x, y, n_train_samples, data

    @abstractmethod
    def _is_dataset(self, file_name, ext):
        """
        Check if a path and extension is to be considered a dataset.

        The check should be simple and only consider the filename and or
        extension of the file. Validation of the file should be deferred to
        `_load_array`

        Parameters
        ----------
        file_name : str
            Name of the dataset file.

        ext : str
            Extension of the dataset file.

        Returns
        -------
        bool
            True if it is a dataset.
        """
        pass

    @abstractmethod
    def _load_array(self, archive, file):
        """
        Load the file inside the archive and convert to a numpy array.

        Parameters
        ----------
        archive : ZipFile
            The zip-archive in which the file reside.
        file : str
            Path to the file inside the zip-archive.

        Returns
        -------
        ndarray
            The dataset converted to a ndarray.
        """
        pass


class NpBundle(Bundle):
    """Bundle of numpy binary files."""

    def _is_dataset(self, file_name, ext):
        return ext in [".npy", ".npz"]

    def _load_array(self, archive, file):
        data = np.load(archive.open(file))

        if isinstance(data, np.ndarray):
            if "x" in self.arrays and "y" in self.arrays:
                return {"x": data[:, :-1], "y": data[:, -1]}
            elif "x" in self.arrays:
                return {"x": data}
            else:
                raise ValueError("Can't infer arrays to export.")
        else:
            return data


class _Dataset:
    def __init__(self, zip_info):
        self.file = zip_info
        self.path, self.ext = os.path.splitext(zip_info.filename)
        self.filename = os.path.basename(self.path)
        if "_TRAIN" in self.filename:
            self.part = "train"
            self.filename = self.filename.replace("_TRAIN", "")
        elif "_TEST" in self.filename:
            self.part = "test"
            self.filename = self.filename.replace("_TEST", "")
        else:
            self.part = "train"


def _sha1(file, buf_size=65536):
    sha1 = hashlib.sha1()
    with open(file, "rb") as f:
        while True:
            data = f.read(buf_size)
            if not data:
                break
            sha1.update(data)
    return sha1.hexdigest()



import numpy as np
from numpy.testing import assert_equal
from wildboar.datasets import load_two_lead_ecg
from wildboar.distance import KMeans, KMedoids, KNeighborsClassifier
from wildboar.utils.estimator_checks import check_estimator


def test_check_estimator():
    check_estimator(KNeighborsClassifier())
    check_estimator(KMeans())
    check_estimator(KMedoids())


def test_kmeans_dtw():
    # fmt: off
    expected = np.array(
        [2, 0, 0, 0, 0, 0, 2, 0, 2, 1, 0, 2, 0, 2, 0, 2, 2, 2, 0, 0, 2, 1,
         2, 2, 0, 0, 0, 0, 2, 1, 0, 1, 0, 0, 0, 0, 2, 1, 0, 0, 2, 1, 0, 1,
         0, 0, 2, 0, 0, 0]
    )
    # fmt: on

    X, y = load_two_lead_ecg()
    k = KMeans(n_clusters=3, random_state=1, metric="dtw")
    k.fit(X[:50])
    assert k.n_iter_ == 5
    assert_equal(k.labels_, expected)


def test_kmeans_euclidean():
    # fmt: off
    expected = np.array(
        [2, 2, 2, 2, 0, 0, 1, 2, 1, 0, 2, 1, 0, 0, 1, 1, 2, 2, 1, 0, 1, 0,
         0, 2, 1, 0, 2, 0, 1, 0, 2, 0, 0, 0, 2, 1, 1, 0, 2, 0, 1, 0, 2, 0,
         1, 1, 1, 0, 2, 2]
    )
    # fmt: on

    X, y = load_two_lead_ecg()
    k = KMeans(n_clusters=3, random_state=1, metric="euclidean")
    k.fit(X[:50])
    assert k.n_iter_ == 5
    assert_equal(k.labels_, expected)



# Authors: Isak Samsten
# License: BSD 3 clause

import numbers

import numpy as np
from sklearn.utils._param_validation import Interval, StrOptions

from ._base import BaseAttributeTransform
from ._crocket import (
    NormalKernelSampler,
    RocketAttributeGenerator,
    ShapeletKernelSampler,
    UniformKernelSampler,
)

_SAMPLING_METHOD = {
    "normal": NormalKernelSampler,
    "uniform": UniformKernelSampler,
    "shapelet": ShapeletKernelSampler,
}


class RocketMixin:
    """
    Mixin for ROCKET based estimators.

    The class provides an implementation for the `_get_generator` method
    with support for rocket convolution.

    The implementing class must have the following properties:

    - `n_kernels`
    - `kernel_size`
    - `min_size`
    - `max_size`
    - `sampling`
    - `sampling_params`
    - `bias_prob`
    - `padding_prob`
    - `normalize_prob`

    See :class:`transform.RocketTransform` for information about the
    properties.
    """

    _parameter_constraints: dict = {
        "n_kernels": [
            Interval(numbers.Integral, 1, None, closed="left"),
        ],
        "kernel_size": ["array-like", None],
        "min_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
            None,
        ],
        "max_size": [
            Interval(numbers.Real, 0, 1, closed="both"),
            None,
        ],
        "sampling": [StrOptions(_SAMPLING_METHOD.keys())],
        "sampling_params": [dict, None],
        "bias_prob": [Interval(numbers.Real, 0, 1, closed="both")],
        "padding_prob": [Interval(numbers.Real, 0, 1, closed="both")],
        "normalize_prob": [Interval(numbers.Real, 0, 1, closed="both")],
    }

    def _get_generator(self, x, y):
        if self.min_size is not None or self.max_size is not None:
            if self.kernel_size is not None:
                raise ValueError(
                    f"Both min_size or max_size and kernel_size of "
                    f"{type(self).__name__} cannot be set at the same time."
                )

            min_size = self.min_size if self.min_size is not None else 0
            max_size = self.max_size if self.max_size is not None else 1

            if min_size > max_size:
                raise ValueError(
                    f"The min_size parameter of {type(self).__name__} "
                    "must be <= max_size."
                )

            max_size = int(self.n_timesteps_in_ * max_size)
            min_size = int(self.n_timesteps_in_ * min_size)
            if min_size < 2:
                min_size = 2
            if max_size < 3:
                max_size = 3

            kernel_size = np.arange(min_size, max_size)
        elif self.kernel_size is None:
            kernel_size = np.array([7, 11, 13], dtype=int)
        else:
            kernel_size = np.array(self.kernel_size, dtype=int)
            if np.min(kernel_size) < 2:
                raise ValueError("The minimum kernel size is 2.")

        KernelSampler = _SAMPLING_METHOD[self.sampling]
        sampling_params = {} if self.sampling_params is None else self.sampling_params
        return RocketAttributeGenerator(
            self.n_kernels,
            KernelSampler(**sampling_params),
            kernel_size,
            self.bias_prob,
            self.padding_prob,
            self.normalize_prob,
        )


class RocketTransform(RocketMixin, BaseAttributeTransform):
    """
    Transform a time series using random convolution features.

    Parameters
    ----------
    n_kernels : int, optional
        The number of kernels to sample at each node.
    sampling : {"normal", "uniform", "shapelet"}, optional
        The sampling of convolutional filters.

        - if "normal", sample filter according to a normal distribution with
          ``mean`` and ``scale``.
        - if "uniform", sample filter according to a uniform distribution with
          ``lower`` and ``upper``.
        - if "shapelet", sample filters as subsequences in the training data.
    sampling_params : dict, optional
        Parameters for the sampling strategy.

        - if "normal", ``{"mean": float, "scale": float}``, defaults to
          ``{"mean": 0, "scale": 1}``.
        - if "uniform", ``{"lower": float, "upper": float}``, defaults to
          ``{"lower": -1, "upper": 1}``.
    kernel_size : array-like, optional
        The kernel size, by default ``[7, 11, 13]``.
    min_size : float, optional
        The minimum timestep size used for generating kernel sizes, If set,
        ``kernel_size`` is ignored.
    max_size : float, optional
        The maximum timestep size used for generating kernel sizes, If set,
        ``kernel_size`` is ignored.
    bias_prob : float, optional
        The probability of using the bias term.
    normalize_prob : float, optional
        The probability of performing normalization.
    padding_prob : float, optional
        The probability of padding with zeros.
    n_jobs : int, optional
        The number of jobs to run in parallel. A value of ``None`` means using
        a single core and a value of ``-1`` means using all cores. Positive
        integers mean the exact number of cores.
    random_state : int or RandomState, optional
        Controls the random resampling of the original dataset.

        - If ``int``, ``random_state`` is the seed used by the random number
          generator.
        - If :class:`numpy.random.RandomState` instance, ``random_state`` is
          the random number generator.
        - If ``None``, the random number generator is the
          :class:`numpy.random.RandomState` instance used by
          :func:`numpy.random`.

    Attributes
    ----------
    embedding_ : Embedding
        The underlying embedding

    References
    ----------
    Dempster, Angus, François Petitjean, and Geoffrey I. Webb.
        ROCKET: exceptionally fast and accurate time series classification using
        random convolutional kernels.
        Data Mining and Knowledge Discovery 34.5 (2020): 1454-1495.

    Examples
    --------
    >>> from wildboar.datasets import load_gun_point
    >>> from wildboar.transform import RocketTransform
    >>> X, y = load_gun_point()
    >>> t = RocketTransform(n_kernels=10, random_state=1)
    >>> t.fit_transform(X)
    array([[0.51333333, 5.11526939, 0.47333333, ..., 2.04712544, 0.24      ,
            0.82912261],
           [0.52666667, 5.26611524, 0.54      , ..., 1.98047216, 0.24      ,
            0.81260641],
           [0.54666667, 4.71210092, 0.35333333, ..., 2.28841158, 0.25333333,
            0.82203705],
           ...,
           [0.54666667, 4.72938203, 0.45333333, ..., 2.53756324, 0.24666667,
            0.8380654 ],
           [0.68666667, 3.80533684, 0.26      , ..., 2.41709413, 0.25333333,
            0.65634235],
           [0.66      , 3.94724793, 0.32666667, ..., 1.85575661, 0.25333333,
            0.67630249]])
    """

    _parameter_constraints: dict = {
        **RocketMixin._parameter_constraints,
        **BaseAttributeTransform._parameter_constraints,
    }

    def __init__(
        self,
        n_kernels=1000,
        *,
        sampling="normal",
        sampling_params=None,
        kernel_size=None,
        min_size=None,
        max_size=None,
        bias_prob=1.0,
        normalize_prob=1.0,
        padding_prob=0.5,
        n_jobs=None,
        random_state=None,
    ):
        super().__init__(n_jobs=n_jobs, random_state=random_state)
        self.sampling = sampling
        self.sampling_params = sampling_params
        self.kernel_size = kernel_size
        self.min_size = min_size
        self.max_size = max_size
        self.bias_prob = bias_prob
        self.normalize_prob = normalize_prob
        self.padding_prob = padding_prob
        self.n_kernels = n_kernels



import numbers

import numpy as np

from ..utils.validation import check_array
from ._cconv import conv1d


def convolve(X, kernel, bias=0.0, *, dilation=1, stride=1, padding=0):
    """
    Apply 1D convolution over a time series.

    Parameters
    ----------
    X : array-like of shape (n_samples, n_timestep)
        The input.
    kernel : array-like of shape (kernel_size, )
        The kernel.
    bias : float, optional
        The bias.
    dilation : int, optional
       The spacing between kernel elements.
    stride : int, optional
        The stride of the convolving kernel.
    padding : int, optional
        Implicit padding on both sides of the input time series.

    Returns
    -------
    ndarray of shape (n_samples, output_size)
        The result of the convolution, where output_size is given by:::

            floor(
                ((X.shape[1] + 2 * padding) - (kernel.shape[0] - 1 * dilation + 1)) / stride
                + 1
            ).

    """
    if not isinstance(dilation, numbers.Integral) or dilation < 1:
        raise ValueError("dilation must be a strictly positive integer.")
    if not isinstance(dilation, numbers.Integral) or stride < 1:
        raise ValueError("stride must be a strictly positive integer.")
    if not isinstance(dilation, numbers.Integral) or padding < 0:
        raise ValueError("padding must be a positive integer.")

    X = check_array(X, allow_3d=False, ensure_ts_array=True, input_name="X")
    kernel = check_array(kernel, ensure_2d=False, order="c", input_name="kernel")

    kernel_size = (kernel.shape[0] - 1) * dilation + 1
    input_size = X.shape[-1] + 2 * padding

    if kernel_size > input_size:
        raise ValueError(f"kernel={kernel_size} larger than input_size={input_size}")

    out = np.empty(
        (X.shape[0], int(np.floor((input_size - kernel_size) / stride) + 1)),
        dtype=float,
    )

    conv1d(
        X,
        kernel,
        bias,
        dilation=int(dilation),
        padding=int(padding),
        stride=int(stride),
        out=out,
    )
    return out



# cython: language_level=3
# cython: cdivision=True
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False

# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np

from libc.math cimport INFINITY, floor, sqrt
from libc.stdlib cimport free, malloc
from libc.string cimport memset
from scipy.linalg.cython_blas cimport dnrm2

from ._fft cimport _pocketfft


cdef double EPSILON = 1e-13

cdef void fast_mean_std(
    const double* data,
    Py_ssize_t length,
    double *mean,
    double* std,
) noexcept nogil:
    """Update the mean and standard deviation"""
    cdef double ex = 0
    cdef double ex2 = 0
    cdef Py_ssize_t i
    for i in range(length):
        current_value = data[i]
        ex += current_value
        ex2 += current_value ** 2

    mean[0] = ex / length
    ex2 = ex2 / length - mean[0] * mean[0]
    if ex2 > EPSILON:
        std[0] = sqrt(ex2)
    else:
        std[0] = 0.0


cdef void inc_stats_init(IncStats *self) noexcept nogil:
    self.mean = 0.0
    self.n_samples = 0.0
    self.sum_square = 0.0
    self.sum = 0.0

cdef void inc_stats_add(IncStats *self, double weight, double value) noexcept nogil:
    cdef double next_m
    self.n_samples += weight
    next_m = self.mean + (value - self.mean) / self.n_samples
    self.sum_square += (value - self.mean) * (value - next_m)
    self.mean = next_m
    self.sum += weight * value

cdef void inc_stats_remove(IncStats *self, double weight, double value) noexcept nogil:
    cdef double old_m
    if self.n_samples == 1.0:
        self.n_samples = 0.0
        self.mean = 0.0
        self.sum_square = 0.0
    else:
        old_m = (self.n_samples * self.mean - value) / (self.n_samples - weight)
        self.sum_square -= (value - self.mean) * (value - old_m)
        self.mean = old_m
        self.n_samples -= weight
    self.sum -= weight * value

cdef double inc_stats_n_samples(IncStats *self) noexcept nogil:
    return self.n_samples

cdef double inc_stats_sum(IncStats *self) noexcept nogil:
    return self.sum

cdef double inc_stats_mean(IncStats *self) noexcept nogil:
    return self.mean

cdef double inc_stats_variance(IncStats *self, bint sample=False) noexcept nogil:
    cdef double n_samples, var
    if sample:
        n_samples = self.n_samples - 1
    else:
        n_samples = self.n_samples
    if n_samples <= 1:
        return 0
    else:
        var = self.sum_square / n_samples
        if var < EPSILON:
            var = 0.0
        return var

cdef void cumulative_mean_std(
    const double *x,
    Py_ssize_t x_length,
    Py_ssize_t y_length,
    double *x_mean,
    double *x_std
) noexcept nogil:
    cdef Py_ssize_t i
    cdef IncStats stats
    inc_stats_init(&stats)
    for i in range(x_length):
        inc_stats_add(&stats, 1.0, x[i])
        if i >= y_length - 1:
            x_mean[i - (y_length - 1)] = stats.mean
            std = sqrt(inc_stats_variance(&stats))
            x_std[i - (y_length - 1)] = std
            inc_stats_remove(&stats, 1.0, x[i - (y_length - 1)])

cdef double mean(const double *x, Py_ssize_t length) noexcept nogil:
    cdef double v = 0.0
    cdef Py_ssize_t i
    for i in range(length):
        v += x[i]
    return v / length

cdef double variance(const double *x, Py_ssize_t length) noexcept nogil:
    if length == 1:
        return 0.0

    cdef double avg = mean(x, length)
    cdef double sum = 0
    cdef double v = 0.0
    cdef Py_ssize_t i
    for i in range(length):
        v = x[i] - avg
        sum += v * v
    return sum / length

cdef double slope(const double *x, Py_ssize_t length) noexcept nogil:
    if length == 1:
        return 0.0
    cdef double y_mean = (length + 1) / 2.0
    cdef double x_mean = 0
    cdef double mean_diff = 0
    cdef double mean_y_sqr = 0
    cdef Py_ssize_t i, j

    for i in range(length):
        j = i + 1
        mean_diff += x[i] * j
        x_mean += x[i]
        mean_y_sqr += j * j
    mean_diff /= length
    mean_y_sqr /= length
    x_mean /= length
    return (mean_diff - y_mean * x_mean) / (mean_y_sqr - y_mean ** 2)

cdef double find_min(
    const double *x, Py_ssize_t n, Py_ssize_t *min_index=NULL
) noexcept nogil:
    cdef double min_val = INFINITY
    cdef Py_ssize_t i

    for i in range(n):
        if x[i] < min_val:
            if min_index != NULL:
                min_index[0] = i
            min_val = x[i]

    return min_val

cdef double covariance(
    const double *x, const double *y, Py_ssize_t length
) noexcept nogil:
    cdef:
        double sum_x = 0.0
        double sum_y = 0.0
        double sum_xy = 0.0
        double k = 0.0
        double mean_x = 0.0
        double mean_y = 0.0
        double tmp_mean_x, tmp_mean_y, diff_x, diff_y
        Py_ssize_t i

    for i in range(length):
        tmp_mean_x = mean_x
        tmp_mean_y = mean_y
        diff_x = x[i] - tmp_mean_x
        diff_y = y[i] - tmp_mean_y
        k += 1
        sum_x += x[i]
        sum_y += y[i]
        sum_xy += diff_x * diff_y - diff_x * diff_y / k
        mean_x = sum_x / k
        mean_y = sum_y / k
    return sum_xy / k


cdef void _auto_correlation(
    const double *x, Py_ssize_t n, double *out, complex *fft
) noexcept nogil:
    cdef double avg = mean(x, n)
    cdef Py_ssize_t fft_length = n * 2 - 1
    cdef Py_ssize_t i
    for i in range(n):
        fft[i] = x[i] - avg
    for i in range(n, fft_length):
        fft[i] = 0.0

    _pocketfft.fft(fft, fft_length, 1.0)
    for i in range(fft_length):
        fft[i] = fft[i] * fft[i].conjugate()

    _pocketfft.ifft(fft, fft_length, 1.0)
    cdef complex first = fft[0]
    if first == 0j:
        first = 1j

    for i in range(n):
        out[i] = (fft[i] / first).real

cdef void auto_correlation(const double *x, Py_ssize_t n, double *out) noexcept nogil:
    cdef Py_ssize_t fft_length = n * 2 - 1
    cdef complex *fft = <complex *> malloc(sizeof(complex) * fft_length)
    _auto_correlation(x, n, out, fft)
    free(fft)

cdef Py_ssize_t next_power_of_2(Py_ssize_t n) noexcept nogil:
    n = n - 1
    while n & n - 1:
        n = n & n - 1
    return n << 1

cdef int welch(
    const double *x,
    Py_ssize_t size,
    int NFFT,
    double Fs,
    double *window,
    int windowWidth,
    double *Pxx,
    double *f,
) noexcept nogil:
    cdef double dt = 1.0 / Fs
    cdef double df = 1.0 / next_power_of_2(windowWidth) / dt
    cdef double m = mean(x, size)
    cdef int k = <int> floor(<double>size / (<double>windowWidth / 2.0)) - 1

    cdef int incx = 1
    cdef double w_norm = dnrm2(&windowWidth, window, &incx)
    cdef double KMU = k * w_norm * w_norm

    cdef double *P = <double*> malloc(sizeof(double) * NFFT)
    cdef complex *F = <complex*> malloc(sizeof(complex) * NFFT)
    cdef double pi
    cdef Py_ssize_t i, j, i_win

    for i in range(NFFT):
        P[i] = 0.0

    for i in range(k):
        i_win = <int>(i * <double>windowWidth/2.0)
        for j in range(windowWidth):
            F[j] = (window[j] * x[j + i_win]) - m

        for j in range(windowWidth, NFFT):
            F[j] = 0.0

        _pocketfft.fft(F, NFFT, 1.0)

        for j in range(NFFT):
            pi = abs(F[j])
            P[j] += pi * pi

    cdef Py_ssize_t n_out = NFFT // 2 + 1
    for i in range(n_out):
        Pxx[i] = P[i] / KMU * dt
        if i > 0 and i < n_out - 1:
            Pxx[i] *= 2
        f[i] = i * df

    free(P)
    free(F)
    return n_out



# Authors: Isak Samsten
# License: BSD 3 clause

import numpy as np
import pytest
import sklearn
from numpy.testing import assert_almost_equal
from sklearn.ensemble import RandomForestClassifier
from sklearn.exceptions import NotFittedError
from sklearn.neighbors import KNeighborsClassifier
from wildboar.base import is_counterfactual
from wildboar.datasets import load_dataset
from wildboar.ensemble import ShapeletForestClassifier
from wildboar.explain.counterfactual import (
    KNeighborsCounterfactual,
    PrototypeCounterfactual,
    ShapeletForestCounterfactual,
    counterfactuals,
)
from wildboar.utils._testing import (
    assert_exhaustive_parameter_checks,
    assert_parameter_checks,
)
from wildboar.utils.estimator_checks import check_estimator

SKLEARN_VERSION = sklearn.__version__.split(".")


@pytest.mark.parametrize(
    "estimator",
    [ShapeletForestClassifier(), KNeighborsCounterfactual(), PrototypeCounterfactual()],
)
def test_estimator_check_shapelet_forest_counterfactual(estimator):
    check_estimator(estimator, skip_scikit=True)


@pytest.mark.parametrize(
    "estimator",
    [
        KNeighborsCounterfactual(),
        PrototypeCounterfactual(),
        ShapeletForestCounterfactual(),
    ],
)
def test_parameter_constraints(estimator):
    assert_exhaustive_parameter_checks(estimator)
    assert_parameter_checks(estimator)


@pytest.mark.parametrize(
    "estimator",
    [
        ShapeletForestCounterfactual(),
        KNeighborsCounterfactual(),
        PrototypeCounterfactual(),
    ],
)
def test_is_counterfactual(estimator):
    assert is_counterfactual(estimator)


@pytest.mark.parametrize(
    "clf, expected_score",
    [
        pytest.param(
            ShapeletForestClassifier(n_shapelets=10, n_estimators=10, random_state=123),
            np.array([1.0216381495890032, 5.925046545971925]),
        ),
        # Some change in scikit-learn > 1.3 changes the
        # output here.
        pytest.param(
            KNeighborsClassifier(n_neighbors=5, metric="euclidean"),
            np.array([2.307078907707276, 8.83732116566978]),
        ),
        pytest.param(
            KNeighborsClassifier(n_neighbors=1, metric="euclidean"),
            np.array([2.6514561178753313, 8.378653689114698]),
        ),
        pytest.param(
            RandomForestClassifier(random_state=123),
            np.array([3.401708, 7.8155463]),
        ),
    ],
)
def test_counterfactuals_best(clf, expected_score):
    x_train, x_test, y_train, y_test = load_dataset(
        "GunPoint", repository="wildboar/ucr-tiny", merge_train_test=False
    )
    clf.fit(x_train, y_train)
    _, _, actual_score = counterfactuals(
        clf,
        x_test[1:3],
        1,
        method="best",
        train_x=x_train,
        train_y=y_train,
        proximity="euclidean",
        random_state=123,
    )
    assert_almost_equal(actual_score, expected_score)


def test_counterfactuals_prototype():
    x_train, x_test, y_train, y_test = load_dataset(
        "GunPoint", repository="wildboar/ucr-tiny", merge_train_test=False
    )
    clf = ShapeletForestClassifier(n_shapelets=10, n_estimators=10, random_state=123)
    clf.fit(x_train, y_train)

    method = PrototypeCounterfactual(method="shapelet", metric="euclidean")
    _, _, actual_score = counterfactuals(
        clf,
        x_test[1:3],
        1,
        method=method,
        train_x=x_train,
        train_y=y_train,
        proximity="euclidean",
        random_state=123,
    )
    expected_score = np.array([2.3592611, 6.2593162])
    assert_almost_equal(actual_score, expected_score)


@pytest.mark.parametrize(
    "counterfactual, estimator",
    [
        (ShapeletForestCounterfactual(), ShapeletForestClassifier(n_shapelets=10)),
        (KNeighborsCounterfactual(), KNeighborsClassifier()),
        (PrototypeCounterfactual(), ShapeletForestClassifier()),
    ],
)
def test_counterfactual_not_fitted_estimator(counterfactual, estimator):
    with pytest.raises(NotFittedError):
        counterfactual.fit(estimator, np.zeros((3, 3)), np.ones(3))



from numpy.testing import assert_almost_equal
from sklearn.metrics import balanced_accuracy_score
from sklearn.model_selection import train_test_split

from wildboar.datasets import load_two_lead_ecg, outlier
from wildboar.ensemble import IsolationShapeletForest
from wildboar.utils.estimator_checks import check_estimator


def test_estimator_checks():
    check_estimator(IsolationShapeletForest())


def load_test_dataset():
    X, y = load_two_lead_ecg()
    X, y = outlier.emmott_outliers(X, y, n_outliers=0.05, random_state=0)
    return train_test_split(X, y, random_state=0)


def test_fit_predict_score_sample():
    X_train, X_test, y_train, y_test = load_test_dataset()
    isf = IsolationShapeletForest(n_shapelets=1, random_state=0)
    isf.fit(X_train)

    expected_score_samples = [
        -0.5346257,
        -0.4549916,
        -0.4623454,
        -0.4922771,
        -0.4988597,
        -0.4394299,
        -0.4887521,
        -0.4621849,
        -0.4910222,
        -0.4955371,
    ]
    assert_almost_equal(isf.score_samples(X_test[:10]), expected_score_samples)
    assert_almost_equal(
        balanced_accuracy_score(y_test, isf.predict(X_test)), 0.9071428571428571
    )



# Authors: Isak Samsten
# License: BSD 3 clause

from joblib import Parallel, delayed, effective_n_jobs


def partition_n_jobs(n_jobs, n):
    batch_size = n // n_jobs
    overflow = n % n_jobs
    offsets = []
    offset = 0
    batch_sizes = []
    for i in range(n_jobs):
        current_overflow = 0
        if i < overflow:
            current_overflow = 1
        current_batch_size = batch_size + current_overflow
        offsets.append(offset)
        batch_sizes.append(current_batch_size)
        offset += current_batch_size

    return n_jobs, offsets, batch_sizes


def run_in_parallel(work, n_jobs=-1, **parallel_args):
    """Execute work in parallel.

    Parameters
    ----------
    work : Batch
        A work batch
    n_jobs : int, optional
        Number of parallel jobs, by default -1
    """
    n_jobs = min(effective_n_jobs(n_jobs), work.n_work)
    if n_jobs == 1:
        work(0, 0, work.n_work)
    else:
        n_jobs, offsets, batch_sizes = partition_n_jobs(n_jobs, work.n_work)
        Parallel(n_jobs=n_jobs, **parallel_args)(
            delayed(work)(jobid, offsets[jobid], batch_sizes[jobid])
            for jobid in range(n_jobs)
        )
